from smartwheel.actions.wheel.base import BaseWheelAction


class WheelAction(BaseWheelAction):
    def __init__(self):
        self.type = "module"

    def run(self, action, canvas, ae):
        if action in [
            "scrollUp",
            "scrollDown",
            "keyAction1",
            "keyAction2",
            "scrollUp2",
            "scrollDown2",
        ]:
            ae().action(action)



import os
import subprocess
import time

from PyQt6.QtCore import *
from PyQt6.QtGui import *

from smartwheel import config, gui_tools
from smartwheel.tools import merge_dicts
from smartwheel.ui.base import BaseUIElem


class ThreadWrapper(QObject):
    update = pyqtSignal(list)


class UIThread(QRunnable):
    def __init__(self, sleep_ms):
        super(UIThread, self).__init__()
        self.wrapper = ThreadWrapper()
        self.unix_meta_old = ""
        self.unix_track_meta_old = []
        self.last_text = []
        self.sleep_ms = sleep_ms
        self.shutdown = False

    @pyqtSlot()
    def run(self):
        while not self.shutdown:
            if os.name == "posix":
                try:
                    res = self.fetchMediaUnix()
                except BaseException as e:
                    print(e)
                    return

            elif os.name == "nt":
                try:
                    res = self.fetchMediaWin()
                except BaseException as e:
                    print(e)
                    return
            else:
                res = ["--", "--", "--", "--"]

            if not res == self.last_text:
                self.wrapper.update.emit(res)
                self.last_text = res

            time.sleep(self.sleep_ms)

    def fetchMediaWin(self):
        return ["--", "--", "--", "--"]

    def findProp(self, s, prop):
        if prop.find(":") != -1:
            exp = lambda x: x[1]
        else:
            exp = lambda x: x[1].split(":")[1]
        for i in s:
            if exp(i) == prop:
                return " ".join(i[2:])
        return "--"

    def fetchMediaUnix(self):
        # TODO move to gi.repository.Playerctl
        raw = subprocess.run(["playerctl", "metadata"], capture_output=True).stdout
        pos = subprocess.run(["playerctl", "position"], capture_output=True).stdout

        if raw == "No players found":
            return ["--", "--", "--", "--"]

        if str(raw) + str(pos) == self.unix_meta_old:
            return self.unix_track_meta_old

        meta = list(
            map(
                lambda x: list([i.decode("utf-8") for i in x.split() if i]),
                raw.splitlines(),
            )
        )

        track = self.findProp(meta, "title")

        if track == "--":
            track = self.findProp(meta, "url").split("/")[-1]

        artist = self.findProp(meta, "artist")
        length = self.findProp(meta, "mpris:length")
        start = "--"
        end = "--"

        if not length == "--":
            length = int(int(length) / 1000000)
            pos = int(float(pos.decode("utf-8")))
            start = time.strftime("%M:%S", time.gmtime(pos))
            end = time.strftime("%M:%S", time.gmtime(length))

        self.unix_meta_old = str(raw) + str(pos)
        self.unix_track_meta_old = [track, artist, start, end]

        return [track, artist, start, end]


class UIElem(BaseUIElem):
    updateSignal = pyqtSignal()

    def __init__(self, config_file, WConfig):
        super().__init__()
        self.config_file = config_file
        self.loadConfig()
        merge_dicts(self.conf, WConfig)
        self.icon_path = self.conf["icon_path"]
        self.thread = UIThread(self.conf["mediaFetchSleep"])
        self.thread.wrapper.update.connect(self.updateText)
        self.text = ["--", "--", "--", "--"]

        self.initGUI()

    def loadConfig(self):
        self.conf = config.Config(self.config_file)
        self.conf.loadConfig()

    def initGUI(self):
        self.updateVars()
        self.pixmap = QPixmap(
            os.path.join(self.conf["iconsFolder"], self.conf["icons"][0])
        )
        self.pixmap = self.pixmap.scaled(
            QSize(self.pix_width, self.pix_height),
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )
        gui_tools.icon_managers["wheel"].colorPixmap(self.pixmap)

    def updateVars(self, offset=0):
        self.width = ((self.conf["width"] + offset) * 3) // 4
        self.height = ((self.conf["height"] + offset) * 3) // 4
        self.pix_width = (self.conf["width"] * 3) // 16
        self.pix_height = (self.conf["height"] * 3) // 16
        self.text_width = ((self.conf["width"] + offset) * 3) // 4
        self.text_height = (self.conf["height"] + offset) // 8

    @pyqtSlot(list)
    def updateText(self, text):
        self.text = text
        self.updateSignal.emit()

    def drawText(self, qp, color, t_font, size, pos, text):
        pen = QPen(QColor(color))
        font = QFont(t_font, size)
        qp.setPen(pen)
        qp.setFont(font)
        qp.drawText(
            QRect(
                self.conf["cx"] - self.text_width // 2,
                self.conf["cy"] - self.text_height // 2,
                self.text_width,
                self.text_height * pos,
            ),
            Qt.AlignmentFlag.AlignCenter,
            text,
        )

    def drawHeaderText(self, qp, t):
        self.drawText(
            qp,
            self.conf["majorTextColor"],
            self.conf["style"]["textHeaderFont"],
            self.conf["style"]["textHeaderSize"],
            1,
            t,
        )

    def drawBottomText(self, qp, t):
        self.drawText(
            qp,
            self.conf["majorTextColor"],
            self.conf["style"]["bottomTextFont"],
            self.conf["style"]["bottomTextSize"],
            2,
            t,
        )

    def drawTrackSeek(self, qp, t1, t2):
        self.drawText(
            qp,
            self.conf["majorTextColor"],
            self.conf["style"]["seekTextFont"],
            self.conf["style"]["seekTextSize"],
            3,
            t1 + " / " + t2,
        )

    def draw(self, qp, offset):
        self.updateVars(offset)
        qp.drawPixmap(
            QPointF(
                self.conf["cx"] - self.pix_width // 2,
                self.conf["cy"] - self.pix_height * 1.5,
            ),
            self.pixmap,
        )

        track, artist, start, end = self.text

        self.drawHeaderText(qp, track)
        self.drawBottomText(qp, artist)
        self.drawTrackSeek(qp, start, end)



import json
import os
import weakref

from PyQt6.QtCore import QObject, pyqtSignal, pyqtSlot

from smartwheel import common


class Config(QObject):
    """
    Universal configuration class that supports loading and saving json config files.

    Supports direct access by key (ConfigObject["key"]) and other dict methods.

    It acts like collections.ChainMap: all source dicts are stored while merging (update method), which is used in updating settings on-the-fly.

    If any property is updated, Config emits `updated()` signal.

    Note: assertion does not work directly! Config = some_other_dict cannot be overloaded, use Config.c = some_other_dict instead.
    """

    updated = pyqtSignal()

    def __init__(
        self,
        config_file=None,
        config_dict=None,
        logger=None,
        ignoreNewVars=True,
        varsWhitelist=[],
        varsBlacklist=[],
        updateFunc=None,
    ):
        """
        Initialize Config object
        Note: the config file needs to be loaded manually by calling loadConfig()

        Parameters
        ==========
        config_file
            (Optional) Path to the configuration file
        config_dict
            (Optional) Dict containing initial values. Overwritten if config_file is specified.
        logger
            logger.Logger instance (recommended, will print to stdout if not given)
        ignoreNewVars
            (Optional) Drop new variables while saving (in order not to write runtime variables to config)
        varsWhitelist
            (Optional) List of keys that override ignoreNewVars option (all new children variables of any depth will be saved)
        varsBlacklist
            (Optional) List of keys that contain runtime variables, new children variables of any depth are ignored
        updateFunc
            (Optional) Update function to call when settings are updated. Use only if signals are not supported
        """
        super(Config, self).__init__()
        self.config_file = config_file
        # self.c = self.loadConfig()
        self.c = config_dict
        self.logger = logger
        self.ignoreNew = ignoreNewVars
        self.whitelist = varsWhitelist
        self.blacklist = varsBlacklist
        self.updateFunc = updateFunc
        self.links = []  # Storing source dicts to allow updating the variables

        common.config_manager.save.connect(self.saveConfig)
        common.config_manager.updated.connect(self.__updated)
        common.config_manager.batchUpdate.connect(self.__batchUpdate)

    def __fetchkey(self, key):
        """
        Get the updated value from the linked dicts. Do not use directly

        Parameters
        ==========
        key
            Dictionary key
        """
        if len(self.links) == 0:
            return self.c[key]

        for i in range(len(self.links)):
            if self.links[i].get(key) is not None:
                return self.links[i][key]
        return self.c[key]

    def update(self, other, include_only=None):
        """
        Call dict update method while saving the linked dict

        Parameters
        ==========
        other
            Other dict to merge
        include_only
            (Optional) Exclude other keys except whitelisted
        """
        if include_only is None:
            self.c.update(other)
        else:
            for key in include_only:
                if other.get(key) is not None:
                    self.c[key] = other[key]

        self.links.append(other)

    def __getitem__(self, key):
        """
        Access config directly (ConfigObject["key"])

        Parameters
        ==========
        key
            The dictionary key
        """
        return self.__fetchkey(key)

    def __setitem__(self, key, newvalue):
        """
        Write to config directly (ConfigObject["key"] = newvalue)

        Parameters
        ==========
        key
            The dictionary key
        newvalue
            Right-hand side value
        """
        self.c[key] = newvalue

    @pyqtSlot(str)
    def __updated(self, key):
        """
        Call the update signal if the property is updated

        Parameters
        ==========
        key
            Updated key
        """
        if self.get(key) is not None:
            if self.updateFunc is not None:
                self.updateFunc()
            self.updated.emit()

    @pyqtSlot(list)
    def __batchUpdate(self, keys):
        """
        Call the update signal if multiple properties may be updated

        Parameters
        ==========
        keys
            Updated keys
        """
        for key in keys:
            if self.get(key) is not None:
                if self.updateFunc is not None:
                    self.updateFunc()
                self.updated.emit()
                break

    def __len__(self):
        return len(self.c)

    def __hash__(self):
        return hash(self.c)

    def __str__(self):
        return str(self.c)

    def __iter__(self):
        return iter(self.c)

    def keys(self):
        return self.c.keys()

    def values(self):
        return self.c.values()

    def get(self, key, default=None):
        """
        Re-implemented dict.get method that respects the linked source dicts

        Parameters
        ==========
        key
            Dict key
        default
            Value to return if the key is not found
        """
        if len(self.links) == 0:
            return self.c.get(key, default)

        for i in range(len(self.links)):
            if self.links[i].get(key) is not None:
                return self.links[i].get(key)

        return self.c.get(key, default)

    def items(self):
        """
        Call the builtin dictionary items method
        """
        return self.c.items()

    def loadConfig(self, immediate=False):
        """
        Load the config from file, returns True if it is loaded successfully

        Parameters
        ==========
        immediate
            If false (default), the function applies the config and does not return it's copy. True - it returns (bool, dict) without applying it
        """
        if self.config_file is None:
            if immediate:
                return True, self.c
            return True

        try:
            with open(self.config_file, "r") as f:
                if immediate:
                    return True, json.load(f)
                self.c = json.load(f)
        except BaseException as e:
            if self.logger is not None:
                self.logger.error("Could not load config file:")
                self.logger.error(str(e))
            else:
                print("Could not load config file:")
                print(str(e))

            if immediate:
                return False, None
            return False

        return True

    def listIter(self, new, old, dropNew=True):
        """
        Recursively iterate through the list and update old config value iff it is present in config file

        Parameters
        ==========
        new
            New values (saved from the application)
        old
            Old values from the json config file
        dropNew
            (Optional) Drop new variables
        """
        if dropNew and not len(new) == len(old):
            return

        # We need to unset the old list in order to prevent duplications errors
        if not dropNew and len(new) < len(old):
            for i in range(len(new), len(old)):
                del old[i]

        for i in range(len(new)):
            if i < len(old):
                if type(new[i]) == type(old[i]):
                    if type(new[i]) == dict:
                        self.dictIter(new[i], old[i], dropNew)
                    elif type(new[i]) == list:
                        self.listIter(new[i], old[i], dropNew)
                    else:
                        old[i] = new[i]
            elif not dropNew:
                old.append(new[i])
            else:
                return

    def dictIter(self, new, old, dropNew=True):
        """
        Recursively iterate through the dict and update old config value iff it is present in config file

        Parameters
        ==========
        new
            New values (saved from the application)
        old
            Old values from the json config file
        dropNew
            (Optional) Drop new variables
        """
        for key, val in new.items():
            if old.get(key) is not None and type(val) == type(old[key]):
                drop = dropNew
                if dropNew and key in self.whitelist:
                    drop = False
                elif not dropNew and key in self.blacklist:
                    drop = True

                if type(val) == dict:
                    self.dictIter(new[key], old[key], drop)
                elif type(val) == list:
                    self.listIter(val, old[key], drop)
                else:
                    old[key] = val
                    # print(key)
            elif not dropNew:
                old[key] = val

    @pyqtSlot()
    def saveConfig(self):
        """
        Save the config file
        Note: new variables are dropped by default (we need to purge runtime variables)
        """
        if self.config_file is None:
            return

        # We need to drop runtime variables, so we need to load the json file again
        ok, old_values = self.loadConfig(immediate=True)

        if not ok:
            if self.logger is not None:
                self.logger.error("Failed to apply config")
            else:
                print("Failed to apply config")

            return

        # recursively iterate over the dictionary
        self.dictIter(self.c, old_values, dropNew=self.ignoreNew)

        with open(self.config_file, "w") as f:
            json.dump(old_values, f, indent=4)



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColor, QTransform

from smartwheel import tools
from smartwheel.backgrounds.base import Background


class PatternBackground(Background):
    def __init__(self, common_config, conf, canvas):
        super(PatternBackground, self).__init__(common_config, conf, canvas)
        tools.merge_dicts(
            self.conf, self.common_config(), include_only=["wheelTextureColor"]
        )

        self.conf.updateFunc = self.draw

        self.pattern_map = {
            "Diagonal (right)": Qt.BrushStyle.BDiagPattern,
            "Grid": Qt.BrushStyle.CrossPattern,
            "Diagonal (left)": Qt.BrushStyle.FDiagPattern,
            "Cross": Qt.BrushStyle.DiagCrossPattern,
            "Vertical": Qt.BrushStyle.VerPattern,
            "Horizontal": Qt.BrushStyle.HorPattern,
            "Dots 1": Qt.BrushStyle.Dense7Pattern,
            "Dots 2": Qt.BrushStyle.Dense6Pattern,
            "None": Qt.BrushStyle.NoBrush,
        }

        self.draw()

    def draw(self):
        self.setColor(QColor(self.conf["wheelTextureColor"]))

        if self.pattern_map.get(self.conf["patternType"]) is not None:
            self.setStyle(self.pattern_map[self.conf["patternType"]])
        else:
            return

        tf = QTransform()
        tf.scale(self.conf["patternScale"], self.conf["patternScale"])
        self.setTransform(tf)


brushes = {"pattern": PatternBackground}



import importlib
import json
import math
import os
import weakref

from PyQt6.QtCore import *
from PyQt6.QtGui import *

from smartwheel.ui.base import BaseUIElem


class UIElem(BaseUIElem):
    def __init__(self):
        pass

    def loadExtensions(self):
        """
        {
            "types": {
                "img": ["jpg", "jpeg"]
            },
            "conf": {
                "img": {
                    "iconPath": "path_to_icon"
                }
            }
        }
        """
        # TODO flatten extensions to "flatten" array
        # TODO generate dict {"type": "common_name"} to "names"
        # TODO load pixmaps under conf -> <type> -> pixmap
        return None

    def get_files(self):
        files = os.listdir(self.conf["baseFolder"])
        ext = self.loadExtensions()
        self.ent = []
        for f in files:
            entity = {}
            entity["filename"] = f
            f_n = f.split(".")
            e = f_n[len(f_n) - 1].lower()
            entity["extension"] = e
            if e in ext["flatten"]:
                entity["type"] = ext["names"][e]
            else:
                entity["type"] = "none"

            entity["conf"] = ext["conf"][entity["type"]]
        self.files_selector = 0
        # TODO generate files selection list (cur_ext variable)
        self.ent.append(entity)

    def draw_file(self, x, y, f):
        self.parent().qp.drawImage(QPoint(x, y), f["conf"]["pixmap"])
        self.parent().qp.drawText(
            (
                x + self.conf["iconWidth"],
                y,
                self.conf["maxTextWidth"],
                self.conf["iconWidth"],
            ),
            Qt.AlignCenter,
            f["filename"],
        )

    def load_selection_icons(self):
        pass

    def draw_selection(self, circleWidth):
        # TODO draw arrow icons (> <)
        top_left = QPoint(
            self.parent().getX(
                self.conf["selectionAngle"], circleWidth - self.conf["iconWidth"]
            )
        )
        bottom_right = QPoint(
            self.parent().getX(
                self.conf["selectionAngle"],
                circleWidth + self.conf["iconWidth"] * 2 + self.conf["maxTextWidth"],
            )
        )
        # TODO set qbrush
        self.qp.drawRect(top_left, bottom_right)
        self.parent().qp.drawImage(
            (
                top_left,
                QPoint(
                    self.parent().getY(
                        self.conf["selectionAngle"],
                        circleWidth - self.conf["iconWidth"],
                    )
                ),
                self.left_arrow_icon,
            )
        )
        self.parent().qp.drawImage(
            (
                bottom_right,
                QPoint(
                    self.parent().getY(
                        self.conf["selectionAngle"],
                        circleWidth + self.conf["iconWidth"],
                    )
                ),
                self.right_arrow_icon,
            )
        )
        # TODO calculate max text width and draw right arrow
        pass

    def draw_files(self, circleWidth):
        for i in range(len(self.cur_ext)):
            self.draw_file(
                self.parent().getX(self.angle * i, circleWidth),
                self.parent().getY(self.angle * i, circleWidth),
                self.cur_ext[i],
            )

    def draw(self, qp, offset=None):
        if self.conf["isWheelWidthFixed"]:
            circleWidth = self.conf["width"] + self.conf["fixedWheelWidth"]
        else:
            circleWidth = (self.conf["width"] * 6) // 5

        # TODO change window size



from smartwheel.actions.baseaction import BaseAction


class Action(BaseAction):
    def __init__(self):
        pass

    def run(self, context):
        event = {
            "canvas": context["canvas"],
            "wheel": context["wheel"],
            "call": context["call"],
        }
        context["module"]().processKey(event)



import json
import os

from PyQt6.QtCore import QObject, pyqtSignal, pyqtSlot

from smartwheel import config


class ConfigManager(QObject):
    """
    Global class that manages all config files. This module contains an instance of this class, acting as a singleton
    """

    save = pyqtSignal()
    updated = pyqtSignal(str)
    batchUpdate = pyqtSignal(list)

    def __init__(self):
        super(ConfigManager, self).__init__()

    def __new__(cls):
        """
        Singleton implementation
        """
        if not hasattr(cls, "instance"):
            cls.instance = super(ConfigManager, cls).__new__(cls)
        return cls.instance

    @pyqtSlot()
    def saveConfig(self):
        """
        Slot function that executes the saving of all Config objects, must be called from settings module
        """
        self.save.emit()


class CacheManager(QObject):
    """
    Global class that manages cache access. Not inteded to be called from other threads
    """

    def __init__(self):
        super(CacheManager, self).__init__()

    def __new__(cls):
        """
        Singleton implementation
        """
        if not hasattr(cls, "instance"):
            cls.instance = super(CacheManager, cls).__new__(cls)
        return cls.instance

    def initManager(self, conf):
        self.conf = conf
        self.conf["cacheDir"] = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), self.conf["cacheDir"]
        )

        if not os.path.exists(self.conf["cacheDir"]):
            os.mkdir(self.conf["cacheDir"])

    def load(self, app_name, filename):
        """
        Load cache file path and config (optional) from cache. Returns (True, filepath, conf) if found

        Parameters
        ==========
        app_name
            Application name, must be unique
        filename
            File to load
        """
        if self.conf is None:
            print("Cache manager is not initialized")
            return False, None, None

        appdir = os.path.join(self.conf["cacheDir"], app_name)
        if not os.path.exists(appdir):
            os.mkdir(appdir)

        cachepath = os.path.join(appdir, filename)
        confpath = os.path.join(appdir, filename.split(".")[0] + ".json")
        cacheconf = None

        if not os.path.exists(cachepath):
            return False, None, None

        if os.path.exists(confpath):
            with open(confpath, "r") as f:
                cacheconf = json.load(f)

        return True, cachepath, cacheconf

    def save(self, app_name, filename, conf=None):
        """
        Save to cache. Returns cache filepath to save manually. Note that config object is saved and loaded automatically

        Parameters
        ==========
        app_name
            Application name, must be unique
        filename
            File to save
        conf
            (Optional) Cache metadata
        """
        if self.conf is None:
            print("Cache manager is not initialized")
            return None

        appdir = os.path.join(self.conf["cacheDir"], app_name)
        if not os.path.exists(appdir):
            os.mkdir(appdir)

        cachepath = os.path.join(appdir, filename)
        confpath = os.path.join(appdir, filename.split(".")[0] + ".json")

        if conf is not None:
            with open(confpath, "w") as f:
                if type(conf) == dict:
                    json.dump(conf, f)
                else:
                    json.dump(conf.c, f)

        return cachepath


config_manager = ConfigManager()
cache_manager = CacheManager()



# from PyQt6.QtGui import *
from PyQt6.QtCore import QObject


class BaseUIElem(QObject):
    def __init__(self):
        super().__init__()
        pass

    def loadConfig(self):
        pass

    def draw(self, qp, offset=None):
        pass



import math

# import importlib
import os
import weakref
from queue import LifoQueue

from PyQt6.QtCore import *
from PyQt6.QtGui import *

from smartwheel import config, gui_tools
from smartwheel.tools import merge_dicts
from smartwheel.ui.base import BaseUIElem


class Section:
    def __init__(self, start_angle, end_angle, parent, module=None):
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.angle = (start_angle + end_angle) // 2
        self.delta = 0
        self.init_angle = parent()._angle
        self.is_selected = False
        self.module = module
        self.parent = parent
        self.pixmap = None
        if self.module is not None:
            self.load_module()
        if self.pixmap is not None:
            self.scale_pixmap()
            self.color_pixmap()

    def load_module(self):
        # mod = importlib.import_module(self.module["name"])
        # ui = mod.UIElem(self.module["config"], self.parent.conf)
        # self.module["class"] = ui
        if self.module["class"].icon_path is None:
            self.pixmap = None  # QImage(os.path.join(self.parent().conf["iconsFolder"], "folder.png"))
        else:
            self.pixmap = QPixmap(
                os.path.join(
                    self.parent().conf["iconsFolder"], self.module["class"].icon_path
                )
            )

    def draw_module(self, qp, opacity):
        if self.module is not None:
            qp.setOpacity(opacity)
            self.module["class"].draw(qp, self.parent()._sections_pos)
            qp.setOpacity(1.0)

    def update_vars(self):
        self.delta = self.parent()._angle - self.init_angle

    def scale_pixmap(self):
        w = self.parent().conf["pixmapScale"]
        self.pixmap.width = w
        self.pixmap.height = w
        self.pixmap = self.pixmap.scaled(
            QSize(w, w),
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )

    def color_pixmap(self):
        gui_tools.icon_managers["sections"].colorPixmap(self.pixmap)

    def reload_module(self, module):
        self.module = module
        self.pixmap = None
        if self.module is not None:
            self.load_module()
        if self.pixmap is not None:
            self.scale_pixmap()

    def draw_icon(self, coords):
        if self.pixmap is not None:
            self.parent().qp.drawPixmap(QPointF(*coords), self.pixmap)

    def calculate_coords(self, a, start, end):
        x1 = self.parent().getX(a, start)
        x2 = self.parent().getX(a, end)
        y1 = self.parent().getY(a, start)
        y2 = self.parent().getY(a, end)
        return x1, x2, y1, y2

    def draw(self, qp, start_w, end_w):
        end_w -= self.parent()._sections_pos
        self.update_vars()
        x1, x2, y1, y2 = self.calculate_coords(
            self.end_angle + self.delta, start_w, end_w
        )

        xa1, xa2, ya1, ya2 = self.calculate_coords(
            self.start_angle + self.delta,
            start_w + self.parent().conf["pixmapScale"],
            end_w,
        )

        xb1, xb2, yb1, yb2 = self.calculate_coords(
            self.end_angle + self.delta,
            start_w + self.parent().conf["pixmapScale"],
            end_w,
        )

        brush = QBrush()
        pen = QPen(
            QColor(self.parent().conf["selectionWheelFG"]), 1, Qt.PenStyle.SolidLine
        )
        qp.setBrush(brush)
        qp.setPen(pen)
        qp.drawLine(QPointF(x1, y1), QPointF(x2, y2))

        self.draw_icon(
            (
                ((xa1 + xa2) // 2 + (xb1 + xb2) // 2) // 2
                - self.parent().conf["pixmapScale"] // 2,
                ((ya1 + ya2) // 2 + (yb1 + yb2) // 2) // 2
                - self.parent().conf["pixmapScale"] // 2,
            )
        )


class UIElem(BaseUIElem):
    def __init__(self, config_file, WConfig, modules, force_update, parent=None):
        super().__init__()
        self.parent = parent
        self.config_file = config_file
        self.modules = modules
        self.force_update = force_update
        self.is_anim_running = False
        self.cur_section = 0
        self.loadConfig()
        merge_dicts(self.conf, WConfig)
        self.initSections()
        self.initAnimation()
        self.initShadowAnimation()
        self.initSectionsAnimation()
        self._opacity = 0
        self._sections_pos = 0
        self.global_shadow = False
        self.wheelUp = LifoQueue()

    def _set_angle(self, a):
        self._angle = a

    def _set_opacity(self, o):
        self._opacity = o

    def _set_sections_pos(self, pos):
        self._sections_pos = pos

    prop_angle = pyqtProperty(int, fset=_set_angle)
    prop_opacity = pyqtProperty(int, fset=_set_opacity)
    prop_sections = pyqtProperty(int, fset=_set_sections_pos)

    def loadConfig(self):
        self.conf = config.Config(self.config_file)
        self.conf.loadConfig()

    def getX(self, a, w):
        return math.cos(math.radians(a)) * w // 2 + self.conf["cx"]

    def getY(self, a, h):
        return math.sin(math.radians(a)) * h // 2 + self.conf["cy"]

    def drawSelection(self, circleWidth, circleHeight):
        # Draw selection wheel
        width = self.conf["width"]
        height = self.conf["height"]

        pen = QPen(QColor(self.conf["selectionWheelFG"]), 1, Qt.PenStyle.SolidLine)
        brush = QBrush(QColor(self.conf["selectionWheelBG"]))
        self.qp.setBrush(brush)
        self.qp.setPen(pen)
        self.qp.drawEllipse(
            QPoint(self.conf["cx"], self.conf["cy"]), width // 2, height // 2
        )

        self.drawSections(width, circleWidth)

    def scrollModule(self, up):
        old_selection = self.cur_section
        if up:
            self.cur_section = (self.cur_section - 1) % len(self.sections)
        else:
            self.cur_section = (self.cur_section + 1) % len(self.sections)
        self.sections[self.cur_section].is_selected = True
        self.sections[old_selection].is_selected = False

    def processKey(self, up):
        # self.scrollModule(up)

        self.wheelUp.put(up)
        if self.is_scroll_anim_running:
            self.updateAnimation(up)
        else:
            self.startAnimation(up)

        self.startShadowAnimation()

        if self.sections_timer.isActive():
            self.sections_timer.start(self.conf["sectionsHideTimeout"])
        else:
            self.sections_timer.start(
                self.conf["sectionsHideTimeout"]
                + self.conf["sectionsAnimationDuration"]
            )
        self.showSections()
        # self.sections_timer.singleShot(self.conf["sectionsHideTimeout"] + self.conf["sectionsAnimationDuration"], self.hideSections)
        # self.startSectionsAnimation(True)

    def openWheel(self):
        self.sections_timer.stop()
        self.showSections()

    def selectModule(self):
        self.sections_timer.stop()
        self.hideSections()

    def quickSwitch(self, up):
        # self.scrollModule(up)
        self.wheelUp.put(up)

        if self.is_scroll_anim_running:
            self.updateAnimation(up)
        else:
            self.startAnimation(up)
        self.startShadowAnimation()

        self.sections_timer.stop()

    def getModule(self, i):
        if i >= len(self.modules):
            return None
        return self.modules[i]

    def initSections(self):
        self.delta = 360 // self.conf["selectionWheelEntries"]
        self._angle = self.conf["selectionAngle"]
        self.sections = [
            Section(
                self.delta * i + self.conf["selectionAngle"] - self.delta // 4,
                self.delta * i
                + self.delta
                + self.conf["selectionAngle"]
                - self.delta // 4,
                weakref.ref(self),
                self.getModule(i),
            )
            for i in range(self.conf["selectionWheelEntries"])
        ]
        self.sections[0].is_selected = True
        self.cur_section = 0

    def resetUI(self):
        for i in range(len(self.sections)):
            self.sections[i].is_selected = False
        self.sections[0].is_selected = True
        self.cur_section = 0

        self.is_sections_hidden = False
        self.sections_timer.stop()

        self.is_scroll_anim_running = False
        self.anim.stop()
        # self.is_shadow_anim_running = False
        # self.shadow_anim.stop()
        self.is_sections_anim_running = False
        self.sections_anim.stop()

        self._angle = self.conf["selectionAngle"]
        self.anim_angle = self._angle
        self._sections_pos = 0
        self._opacity = 0
        self.wheelUp = LifoQueue()

        self.global_shadow = True
        self.resetShadowAnimation()
        self.startShadowAnimation()

    def reloadModules(self, modules):
        self.resetUI()
        self.modules = modules
        for i in range(len(self.sections)):
            self.sections[i].reload_module(self.getModule(i))

    def getSectionsModules(self):
        return [x.module for x in self.sections]

    def getCurModule(self):
        return self.cur_section

    def initAnimation(self):
        self.is_scroll_anim_running = False
        self.anim_angle = self.conf["selectionAngle"]
        self.anim_start = self.anim_angle
        self.anim_end = self.anim_angle + self.delta
        self.anim = QPropertyAnimation(self, b"prop_angle")
        self.anim.setDuration(self.conf["scrollAnimationDuration"])
        self.anim.setEasingCurve(QEasingCurve.Type.InOutQuad)

    def startAnimation(self, up, a=None):
        self.is_scroll_anim_running = True
        if a is not None:
            self.anim_start = a
        else:
            self.anim_start = self.anim_angle
        if up:
            self.anim_angle += self.delta
            self.anim_end = self.anim_angle
        else:
            self.anim_angle -= self.delta
            self.anim_end = self.anim_angle
        self.anim.setStartValue(self.anim_start)
        self.anim.setEndValue(self.anim_end)
        self.anim.start()

    def updateAnimation(self, up):
        self.anim.stop()
        self.startAnimation(up, a=self._angle)

    def drawShadow(self, cw):
        if self._opacity == 0:
            return
        pen = QPen(QColor(self.conf["wheelTextureColor"]))
        self.qp.setPen(pen)
        opac = self._opacity / 255  # [0.0, 1.0]

        # self.qp.setCompositionMode(QPainter.CompositionMode.CompositionMode_Destination)

        self.qp.setOpacity(opac)
        brush = QBrush(QColor(self.conf["bgWheelColor"]))
        self.qp.setBrush(brush)
        self.qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), cw // 2, cw // 2)

        # brush = self.parent().brushes.get(self.conf["backgroundStyle"])
        # if brush is None:
        brush = QBrush(
            QColor(self.conf["wheelShadowColor"]),
        )

        self.qp.setBrush(brush)
        self.qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), cw // 2, cw // 2)

        self.qp.setOpacity(1.0)
        # self.qp.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceOver)

    def initShadowAnimation(self):
        self.is_shadow_anim_running = False
        self.shadow_anim = QPropertyAnimation(self, b"prop_opacity")
        self.shadow_anim.setDuration(self.conf["shadowAnimationDuration"])
        self.resetShadowAnimation()

    def resetShadowAnimation(self, up=True):
        if up:
            self.shadow_anim_start = 0
            self.shadow_anim_end = 255
        else:
            self.shadow_anim_start = 255
            self.shadow_anim_end = 0

    @pyqtSlot()
    def shadowAnimationMiddle(self):
        while not self.wheelUp.empty():
            self.scrollModule(self.wheelUp.get())

        self.resetShadowAnimation(False)
        self.shadow_anim.finished.disconnect()
        self.startShadowAnimation(False)

    def startShadowAnimation(self, up=True):
        dur = self.conf["shadowAnimationDuration"]
        if self.is_shadow_anim_running == True:
            # if not self.wheelUp.empty():
            #    self.scrollModule(self.wheelUp.get())
            self.shadow_anim.stop()
            self.shadow_anim_start = self._opacity
            dur = dur * (1 - abs(self.shadow_anim_start - self._opacity) / 255)

        self.shadow_anim.setStartValue(self.shadow_anim_start)
        self.shadow_anim.setEndValue(self.shadow_anim_end)
        # self.shadow_anim.setEndValue(self.shadow_anim_end)
        self.shadow_anim.setDuration(int(dur) // 2)
        if up:
            self.shadow_anim.finished.connect(self.shadowAnimationMiddle)
        self.shadow_anim.start()
        self.is_shadow_anim_running = True
        self.resetShadowAnimation()

    def initSectionsAnimation(self):
        self.is_sections_anim_running = False
        self.is_sections_hidden = False
        self.sections_anim_start = 0
        self.sections_timer = QTimer()
        self.sections_timer.setSingleShot(True)
        # self.sections_timer.setInterval(self.conf["sectionsHideTimeout"])
        # self.sections_timer.singleShot(self.conf["sectionsHideTimeout"], self.hideSections)
        self.sections_timer.timeout.connect(self.hideSections)
        self.sections_timer.start(self.conf["sectionsHideTimeout"])
        if self.conf["isWheelWidthFixed"]:
            self.sections_anim_end = self.conf["fixedWheelWidth"]
        else:
            self.sections_anim_end = (self.conf["width"] * 1) // 4
        self.sections_anim = QPropertyAnimation(self, b"prop_sections")
        self.sections_anim.setDuration(self.conf["sectionsAnimationDuration"])

    def startSectionsAnimation(self, hide):
        self.is_sections_anim_running = True
        if hide:
            self.sections_anim.setStartValue(self._sections_pos)
            self.sections_anim.setEndValue(self.sections_anim_end)
            self.is_sections_hidden = True
        else:
            self.sections_anim.setStartValue(self._sections_pos)
            self.sections_anim.setEndValue(self.sections_anim_start)
            self.is_sections_hidden = False
        self.sections_anim.start()
        self.force_update()

    def hideSections(self):
        if not self.is_sections_hidden:
            self.startSectionsAnimation(True)

    def showSections(self):
        if self.is_sections_hidden:
            self.startSectionsAnimation(False)

    def drawSections(self, w, cw):
        for s in self.sections:
            s.draw(self.qp, w, cw)

        """
        self.qp.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)
        color = QColor(self.conf["iconColor"])
        self.qp.setBrush(color)
        self.qp.setPen(color)
        self.qp.drawRect(self.conf["corner_x"], self.conf["corner_y"], self.conf["width"], self.conf["height"])
        for s in self.sections:
            s.draw_icon()
        """

        pen = QPen(QColor(self.conf["pointerColor"]), 3, Qt.PenStyle.SolidLine)
        self.qp.setPen(pen)
        # draw pointer
        self.qp.drawArc(
            self.conf["corner_x"] - self.conf["pointerMargin"],
            self.conf["corner_y"] - self.conf["pointerMargin"],
            self.conf["width"] + self.conf["pointerMargin"],
            self.conf["height"] + self.conf["pointerMargin"],
            -90 * 16 + self.conf["selectionAngle"] * 16 + (self.delta // 4) * 16,
            -self.delta * 16,
        )
        # p = QPalette()
        # gradient = QConicalGradient(self.conf["cx"], self.conf["cy"], self.conf["selectionAngle"])
        # gradient.setColorAt(0.0, QColor.fromHsv(0, 0, 255))
        # gradient.setColorAt(1.0, QColor.fromHsv(255, 0, 255))
        # self.qp.setBrush(QBrush(gradient))
        # self.qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), cw // 3, cw // 3)

    def drawMainWheel(self, circleWidth, circleHeight):
        # Draw main wheel
        color = self.conf["bgWheelColor"]
        if self.conf["drawWheelCircle"]:
            color = self.conf["wheelTextureColor"]
        pen = QPen(QColor(color), 1, Qt.PenStyle.SolidLine)
        self.qp.setPen(pen)
        brush = QBrush(QColor(self.conf["bgWheelColor"]))
        self.qp.setBrush(brush)

        self.qp.drawEllipse(
            QPoint(self.conf["cx"], self.conf["cy"]),
            circleWidth // 2,
            circleHeight // 2,
        )

        # self.qp.setOpacity(1.0 - self._opacity / 255)
        brush = self.parent().brushes.get(self.conf["backgroundStyle"])
        if brush is None:
            brush = QBrush(
                QColor(self.conf["wheelTextureColor"]), Qt.BrushStyle.BDiagPattern
            )

        self.qp.setBrush(brush)
        self.qp.drawEllipse(
            QPoint(self.conf["cx"], self.conf["cy"]),
            circleWidth // 2,
            circleHeight // 2,
        )
        # self.qp.setOpacity(1.0)
        self.drawOverlays(1 - self._opacity / 255, circleWidth)

    def drawOverlays(self, opacity, circleWidth):
        brush = QBrush(QColor(self.conf["overlayCirclesColor"]))

        self.qp.setBrush(brush)

        if self.conf["drawOverlayCircles"]:
            self.qp.setPen(QPen(QColor(self.conf["overlayCirclesColor"])))
            self.qp.setOpacity(opacity * self.conf["overlayCirclesOpacity"])

            for i in range(0, 360, 360 // 10):
                self.qp.drawEllipse(
                    QPoint(
                        self.conf["cx"]
                        + int(
                            math.cos(math.radians(i + (self._angle - 225) / 4))
                            * circleWidth
                            // 3
                        ),
                        self.conf["cy"]
                        + int(
                            math.sin(math.radians(i + (self._angle - 225) / 4))
                            * circleWidth
                            // 3
                        ),
                    ),
                    self.conf["overlayCirclesWidth"],
                    self.conf["overlayCirclesWidth"],
                )
        if self.conf["drawOverlayRects"]:
            self.qp.setPen(
                QPen(
                    QColor(self.conf["overlayRectsColor"]),
                    self.conf["overlayRectsWidth"],
                    Qt.PenStyle.SolidLine,
                )
            )
            self.qp.setOpacity(self.conf["overlayRectsOpacity"] * opacity)

            for i in range(0, 360, 360 // 6):
                cx = math.cos(math.radians(i + (self._angle - 225) / 6))
                cy = math.sin(math.radians(i + (self._angle - 225) / 6))
                self.qp.drawLine(
                    QPointF(
                        self.conf["cx"] + cx * circleWidth * 0.3,
                        self.conf["cy"] + cy * circleWidth * 0.3,
                    ),
                    QPointF(
                        self.conf["cx"] + cx * circleWidth * 0.45,
                        self.conf["cy"] + cy * circleWidth * 0.45,
                    ),
                )

        self.qp.setOpacity(1.0)

    def draw(self, qp, offset=None):
        if self.anim.state() == QAbstractAnimation.State.Stopped:
            self.is_scroll_anim_running = False
        if self.shadow_anim.state() == QAbstractAnimation.State.Stopped:
            self.global_shadow = False
            self.is_shadow_anim_running = False
        if self.sections_anim.state() == QAbstractAnimation.State.Stopped:
            self.is_sections_anim_running = False
        self.is_anim_running = (
            self.is_scroll_anim_running
            or self.is_shadow_anim_running
            or self.is_sections_anim_running
        )
        if self.conf["isWheelWidthFixed"]:
            circleWidth = (
                self.conf["width"] - self.conf["fixedWheelWidth"] + self._sections_pos
            )
            circleHeight = (
                self.conf["height"] - self.conf["fixedWheelWidth"] + self._sections_pos
            )
        else:
            circleWidth = (self.conf["width"] * 3) // 4 + self._sections_pos
            circleHeight = (self.conf["height"] * 3) // 4 + self._sections_pos

        self.qp = qp

        self.drawSelection(circleWidth, circleHeight)

        self.drawMainWheel(circleWidth, circleHeight)

        self.sections[self.cur_section].draw_module(self.qp, 1.0 - self._opacity / 255)

        if self.global_shadow:
            self.drawShadow(self.conf["width"])
        # else:
        #    self.drawShadow(circleWidth)

        self.drawOverlays(self._opacity / 255, circleWidth)



from PyQt6.QtCore import QObject, pyqtSlot


class BaseHandler(QObject):
    """
    Base settings handler class
    """

    def __init__(self, value_getter, value_setter, parent_obj=None):
        """
        Initialize the handler

        Parameters
        ==========
        value_getter
            Property getter function (settings.SettingsWindow.getValue)
        value_setter
            Property setter function (settings.SettingsWindow.setValue)
        parent_object
            Weakref to the parent object
        """
        super(BaseHandler, self).__init__()
        self.value_getter = value_getter
        self.value_setter = value_setter
        self.parent_obj = parent_obj

    def initElem(self, elem):
        """
        Initialize elem with specified type, must return QWidget

        Parameters
        ==========
        elem
            Element dictionary (loaded from the settings registry)
        """
        pass

    @pyqtSlot()
    def setter(self):
        """
        Setter slot, needs to be manually connected to the related widget
        """
        pass

    def fetchValue(self, wid):
        """
        (Presets) Get value of the widget, must return the value or None if not supported

        Parameters
        ==========
        wid
            Widget to fetch
        """
        return None

    def updateValue(self, wid, value):
        """
        (Presets) Set the value of the widget, must return True if ok, False if not supported

        Parameters
        ==========
        wid
            Widget to update
        value
            Value to set
        """
        return False

    def linkElem(self, elem, registriesName):
        """
        (External settings registries) Link element to settings group "registriesName"

        Widget signal must be linked to settings.showLinkedWidgets(str), where str is the selected element name

        Function must return True if supported

        Parameters
        ==========
        elem
            Widget to link
        registriesName
            Widget property to set
        """
        return False


handlers = {}



import json
import logging
import os
import weakref

from PyQt6.QtCore import pyqtSlot
from PyQt6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QInputDialog,
    QLineEdit,
    QPushButton,
    QWidget,
)

from smartwheel.settings_handlers.base import BaseHandler


class PresetHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(PresetHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        layout = QHBoxLayout()
        wid = QComboBox()
        ok, self.basedir = self.value_getter(module="canvas", prop="basedir")

        wid.setProperty("presets_folder", elem["folder"])

        p_path = os.path.join(self.basedir, "presets", elem["folder"])

        if not os.path.exists(p_path):
            os.makedirs(p_path, exist_ok=True)

        preset_files = [x for x in os.listdir(p_path) if x.endswith(".json")]

        presets = {}
        preset_names = []

        for p in preset_files:
            with open(
                os.path.join(self.basedir, "presets", elem["folder"], p), "r"
            ) as f:
                pr = json.load(f)
                presets[pr["name"]] = pr

        for _, p in presets.items():
            preset_names.append(p["title"])

        preset_names.append("Custom")

        self.value_setter(
            value=presets, module="settings", prop="presets." + str(elem["tab_index"])
        )

        wid.insertItems(len(preset_names) - 1, preset_names)
        wid.setProperty("tab_index", elem["tab_index"])
        wid.setProperty("module", elem["module"])
        wid.setProperty("prop", elem["prop"])

        wid.currentIndexChanged.connect(self.setPreset)

        ok, value = self.value_getter(elem["module"], elem["prop"])
        if not ok:
            value = "Custom"

        wid.setCurrentText(value)

        save = QPushButton("Save")
        save.setProperty("combo", weakref.ref(wid))
        save.clicked.connect(self.savePreset)

        layout.addWidget(wid)
        layout.addWidget(save)

        wrapper = QWidget()
        wrapper.setLayout(layout)
        return wrapper

    def parsePresetName(self, title):
        """
        Convert preset title to the internal preset name ("Arctic Flat" -> "arctic_flat")

        Parameters
        ==========
        title
            Title to convert
        """
        return "".join(
            [
                x if x.isalnum or x == "_" else ""
                for x in title.lower().replace(" ", "_")
            ]
        )

    @pyqtSlot()
    def savePreset(self):
        """
        Save current preset, must be called by the Save button
        """
        caller = self.sender().property("combo")
        if caller is None:
            self.logger.warning("Could not get linked combo box")
            return

        ind = caller().property("tab_index")
        folder = caller().property("presets_folder")
        module = caller().property("module")
        prop = caller().property("prop")

        if ind is None:
            self.logger.warning("Could not get tab index from the preset picker")
            return
        if folder is None:
            self.logger.warning(
                "Could not get presets folder name from the preset picker"
            )
            return
        if module is None or prop is None:
            self.logger.warning("Could not get presets picker properties")
            return

        preset = caller().currentText()

        if preset == "Custom":
            title, ok = QInputDialog().getText(
                self.sender(),
                "New preset",
                "Preset name:",
                QLineEdit.EchoMode.Normal,
                "Preset" + str(caller().currentIndex() + 1),
            )

            if not ok or not title:
                self.logger.info("Input dialog closed")
                return

            if caller().findText(title) == -1:
                new_ind = caller().count() - 1
                caller().insertItem(new_ind, title)
                caller().blockSignals(True)
                caller().setCurrentIndex(new_ind)
                caller().blockSignals(False)
                self.value_setter(value=title, module=module, prop=prop)
        else:
            title = preset

        name = self.parsePresetName(title)

        self.parent_obj().savePreset(
            ind,
            name,
            title,
            os.path.join(self.basedir, "presets", folder, name + ".json"),
        )

    @pyqtSlot(int)
    def setPreset(self, value):
        """
        Load the preset from the combo

        Parameters
        ==========
        value
            Combo index
        """
        caller = self.sender()
        ind = caller.property("tab_index")
        module = caller.property("module")
        prop = caller.property("prop")

        if ind is None:
            self.logger.warning("Could not get tab index from the preset picker")
        if module is None or prop is None:
            self.logger.warning("Could not get presets picker properties")
            return

        preset = caller.currentText()

        self.value_setter(value=preset, module=module, prop=prop)

        if preset == "Custom":
            return

        self.parent_obj().loadPreset(ind, self.parsePresetName(preset))


handlers = {"preset": PresetHandler}



import config
from PyQt6.QtCore import *

from smartwheel.serialpipe.base import ConnPipe


class SConn(ConnPipe):
    """Control the wheel from keypresses (no background scanner)"""

    def __init__(self, config_file, call_signal):
        """
        Initialize keyboardpipe

        Parameters
        ----------
        config_file
            Configuration file
        call_signal
            Actionegnine signal to call
        """
        super().__init__()
        self.conf = None
        self.config_file = config_file
        self.call = call_signal
        self.loadConfig()

    def loadConfig(self):
        self.conf = config.Config(self.config_file)
        self.conf.loadConfig()

    def findKey(self, key):
        """
        Returns command tuple by key

        Parameters
        ----------
        key
            String containing the key
        """
        for b in self.conf["binds"]:
            for c in b["commands"]:
                if c["string"] == key:
                    return b, c

        return None

    def handleKeypress(self, event):
        """
        Keypress handler, called from main thread
        """
        if event.key() == Qt.Key_W:
            self.call.emit(self.findKey("w"))
        elif event.key() == Qt.Key_A:
            self.call.emit(self.findKey("a"))
        elif event.key() == Qt.Key_Up:
            self.call.emit(self.findKey("up"))
        elif event.key() == Qt.Key_Down:
            self.call.emit(self.findKey("down"))
        elif event.key() == Qt.Key_Return:
            self.call.emit(self.findKey("return"))
        elif event.key() == Qt.Key_Escape:
            self.call.emit(self.findKey("esc"))

    def run(self):
        """
        We don't need the run loop, since we are reading keyboard events straight from Qt
        """
        pass



import importlib
import logging
import os
import weakref

from PyQt6.QtCore import *

from smartwheel import config, tools


class ActionEngine(QObject):
    """Modules and wheel interactions interface"""

    callAction = pyqtSignal(tuple, name="action_call")

    def __init__(self, modules, config_file, WConfig):
        """
        Initialize ActionEngine

        Parameters
        ----------
        modules
            Wheel modules
        config_file
            Configuration file (actionengine.json)
        WConfig
            Canvas config
        """
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.importConfig(config_file)
        tools.merge_dicts(self.conf, WConfig)
        self.current_module = 0
        self.current_module_getter = None
        self.current_module_list_getter = None
        self.canvas = None
        self.wheel = None
        self.modules = modules
        self.callAction.connect(self.processCall)
        self.actions = {}
        self.importActions()
        self.importWheelActions()

    def importConfig(self, config_file):
        """
        Load config file (actionengine.json)

        Parameters
        ----------
        config_file
            Filename

        """
        self.conf = config.Config(
            config_file, self.logger, varsWhitelist=["commandBind"]
        )
        self.conf.loadConfig()

    def importActions(self):
        """
        Import actions from `actions` folder
        """
        for mod in self.conf["actionModulesLoad"]:
            name = self.conf["actionModules"][mod]["name"]
            self.actions[name.split(".")[-1:][0]] = importlib.import_module(
                "smartwheel." + name
            ).Action()

    def importWheelActions(self):
        """
        Import actions from `actions/wheel` folder
        """
        self.wheel_actions = {}
        for mod in self.conf["wheelActionModulesLoad"]:
            mod_class = importlib.import_module(
                "smartwheel." + self.conf["wheelActionModules"][mod]["name"]
            ).WheelAction()
            self.wheel_actions[mod_class.type] = mod_class

    def getModule(self, i):
        """
        Get i-th module

        Parameters
        ----------
        i
            Index of module
        """
        if i >= len(self.modules):
            return None
        return self.modules[i]

    def action(self, call):
        """
        Execute action

        Parameters
        ----------
        call
            Module action (from actions list)
        """
        self.modules = self.current_module_list_getter()
        self.current_module = self.current_module_getter()
        if self.getModule(self.current_module) == None:
            return
        context = (
            self.modules[self.current_module]["class"].conf["actions"].get(call, None)
        )
        if context is None:
            return
        for i in context:
            i["canvas"] = self.canvas
            i["wheel"] = self.wheel
            i["module"] = weakref.ref(self.modules[self.current_module]["class"])
            i["call"] = call
            self.actions[i["action"].lower()].run(i)

    def getWheelAction(self, a):
        """
        Find wheel action
        a
           TODO understand this
        """
        for i in self.conf["commandActions"]:
            if i["type"] == a["mode"] and i["name"] == a["action"]:
                return i
        return None

    def getState(self):
        """
        Get current state (sections opened or closed)
        """
        if self.canvas().conf["modules"][0]["class"].is_sections_hidden:
            return "module"
        return "wheel"

    @pyqtSlot(tuple)
    def processCall(self, p_call):
        """
        Execute action by call (slot function)
        This is the main interface of ActionEngine, check serial.py for example

        Parameters
        ----------
        p_call
            callAction in the form of (bind, command)
            bind is a dict that contains 'name' string
            command is a dict with 'string' field (the command)
        """
        elem, call = p_call
        cur_state = self.getState()

        if elem.get("name") is None or call.get("string") is None:
            self.logger.warning("actionengine could not parse the signal")

        self.logger.debug("Incoming call: " + call["string"])

        i = self.conf["commandBind"].get(elem["name"], None)
        if i is not None:
            c = list(j for j in i if j["command"] == call["string"])
            if c:  # c != []
                for act in c:
                    for a in act["actions"]:
                        cmd = self.getWheelAction(a)
                        if cmd is not None:
                            onState = a.get("onState", None)
                            if onState is not None:
                                if a.get("checkState", True) and onState != cur_state:
                                    continue
                            elif a.get("checkState", True) and a["mode"] != cur_state:
                                continue
                            for _ in range(1 + a.get("repeat", 0)):
                                self.wheel_actions[cmd["type"]].run(
                                    cmd["name"], self.canvas, weakref.ref(self)
                                )
                self.canvas().update_func()
        else:
            self.logger.warning("actionengine could not find call with this name")

    def loadModulesConf(self, conf):
        self.modules.append(conf)



import logging

from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from smartwheel.settings_handlers.base import BaseHandler


class ActionPicker(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(ActionPicker, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)
        self.pickers = []  # Prevent pickers from being destroyed

    def initElem(self, elem):
        """
        Initialize actions picker, should not be called directly from settings.py

        Parameters
        ==========
        elem
            Action engine command {"name": "commandTitle", "device": "commandBind", "command": "commandName"}
        """
        wrapper = QWidget()
        wid = QHBoxLayout()
        sp = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        label = QLabel(elem["name"])
        editButton = QPushButton("...")
        deleteButton = QPushButton("x")

        # Fetching the ActionList
        if self.parent_obj is None:
            self.logger.error("Could not get settings object")
            return None

        picker = self.parent_obj().handlers["actions_list"].initElem(elem)
        if picker is None:
            self.logger.error("Could not initialize actions window")
            return None

        self.pickers.append(picker)

        editButton.clicked.connect(picker.show)

        wid.addWidget(label)
        wid.addSpacerItem(sp)
        wid.addWidget(editButton)
        wid.addWidget(deleteButton)

        wrapper.setLayout(wid)

        return wrapper


class OnShown(QWidget):
    """
    Widget that calls the shown signal when visible
    Qt has no proper way to have 2 gridlayouts with the same column size, so I had to implement this hack
    """

    shown = pyqtSignal()

    def showEvent(self, event):
        self.shown.emit()
        event.accept()


class ActionList(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(ActionList, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        """
        Initialize actions picker window, should not be called directly from settings.py

        Parameters
        ==========
        elem
            Action engine command name {"device": "commandBind", "command": "commandName"}
        """
        # if parent_obj is None:
        #    self.logger.error("Could not load action editor: no parent object found")
        #    return None
        wrapper = QWidget()
        wrapper.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        layout = QVBoxLayout()

        wheel_l = QGridLayout()
        module_l = QGridLayout()

        wheel = QGroupBox("Wheel actions")
        module = QGroupBox("Module actions")

        ok, actions = self.value_getter("actionengine", "commandActions")

        if not ok:
            self.logger.error("Could not get actionengine config")
            return None

        if elem.get("device") is None or elem.get("command") is None:
            self.logger.error(
                "No command bind specified, could not launch action editor"
            )
            return None

        # {"command": "commandName", "actions": [{"action": "...", "mode": "...", "repeat": ..}, ...]}
        ok, elem_bind = self.value_getter(
            "actionengine", "commandBind." + elem["device"]
        )
        if not ok:
            elem_bind = []

        elem_actions = []
        for e in elem_bind:
            if e.get("command") is not None and e["command"] == elem["command"]:
                elem_actions = e.get("actions", [])

        font = QFont()
        font.setBold(True)

        for i, s in enumerate([10, 0, 0, 0]):
            wheel_l.setColumnStretch(s, i)
            module_l.setColumnStretch(s, i)

        for group in ["wheel", "module"]:
            enabled_label = QLabel("Enabled")
            title_label = QLabel("Title")
            any_label = QLabel("Where to call")

            desc_label = QLabel("Description")
            desc_label.setFont(font)

            enabled_label.setFont(font)
            title_label.setFont(font)
            any_label.setFont(font)

            if group == "wheel":
                wheel_l.addWidget(enabled_label, 0, 0, Qt.AlignmentFlag.AlignLeft)
                wheel_l.addWidget(title_label, 0, 1, Qt.AlignmentFlag.AlignLeft)
                wheel_l.addWidget(desc_label, 0, 2, Qt.AlignmentFlag.AlignLeft)
                wheel_l.addWidget(any_label, 0, 3, Qt.AlignmentFlag.AlignLeft)

            if group == "module":
                module_l.addWidget(enabled_label, 0, 0, Qt.AlignmentFlag.AlignLeft)
                module_l.addWidget(title_label, 0, 1, Qt.AlignmentFlag.AlignLeft)
                module_l.addWidget(desc_label, 0, 2, Qt.AlignmentFlag.AlignLeft)
                module_l.addWidget(any_label, 0, 3, Qt.AlignmentFlag.AlignLeft)

        for i, a in enumerate(actions):
            enabled = QCheckBox()

            enabled.setProperty("action_name", a["name"])

            title = QLabel(a["title"])
            desc = QLabel(a["description"])

            state = QComboBox()
            state.insertItems(0, ["Wheel", "Module", "Anywhere"])
            # anyState = QCheckBox()

            if a.get("default") is not None:
                if a["default"] == "any":
                    state.setCurrentIndex(2)
                elif a["default"] == "wheel":
                    if bind["mode"] == "module":
                        state.setCurrentIndex(1)
                    else:
                        state.setCurrentIndex(0)
                elif a["default"] == "module":
                    if bind["mode"] == "wheel":
                        state.setCurrentIndex(0)
                    else:
                        state.setCurrentIndex(1)

            for j, bind in enumerate(elem_actions):
                if bind["action"] == a["name"]:
                    enabled.setChecked(True)
                else:
                    continue

                if a["type"] == "wheel":
                    if bind.get("onState") is not None and bind["onState"] == "module":
                        state.setCurrentIndex(1)
                    else:
                        state.setCurrentIndex(0)
                if a["type"] == "module":
                    if bind.get("onState") is not None and bind["onState"] == "wheel":
                        state.setCurrentIndex(0)
                    else:
                        state.setCurrentIndex(1)

                if not bind["checkState"]:
                    state.setCurrentIndex(2)

            if a["type"] == "wheel":
                wheel_l.addWidget(enabled, i + 1, 0, Qt.AlignmentFlag.AlignLeft)
                wheel_l.addWidget(title, i + 1, 1, Qt.AlignmentFlag.AlignLeft)
                wheel_l.addWidget(desc, i + 1, 2, Qt.AlignmentFlag.AlignLeft)
                wheel_l.addWidget(state, i + 1, 3, Qt.AlignmentFlag.AlignRight)

                for j in range(1, 4):
                    if wheel_l.itemAtPosition(i + 1, j) is not None:
                        enabled.clicked.connect(
                            wheel_l.itemAtPosition(i + 1, j).widget().setEnabled
                        )
                        if not enabled.isChecked():
                            wheel_l.itemAtPosition(i + 1, j).widget().setDisabled(True)
            else:
                module_l.addWidget(enabled, i + 1, 0, Qt.AlignmentFlag.AlignLeft)
                module_l.addWidget(title, i + 1, 1, Qt.AlignmentFlag.AlignLeft)
                module_l.addWidget(desc, i + 1, 2, Qt.AlignmentFlag.AlignLeft)
                module_l.addWidget(state, i + 1, 3, Qt.AlignmentFlag.AlignRight)

                for j in range(1, 4):
                    enabled.clicked.connect(
                        module_l.itemAtPosition(i + 1, j).widget().setEnabled
                    )
                    if not enabled.isChecked():
                        module_l.itemAtPosition(i + 1, j).widget().setDisabled(True)

        panel = QHBoxLayout()
        okButton = QPushButton("OK")
        okButton.clicked.connect(self.applyChanges)
        okButton.clicked.connect(wrapper.close)

        cancelButton = QPushButton("Cancel")
        cancelButton.clicked.connect(wrapper.close)

        onShown = OnShown()
        onShown.shown.connect(self.resizeCells)
        panel.addWidget(onShown)

        spacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        panel.addSpacerItem(spacer)
        panel.addWidget(cancelButton)
        panel.addWidget(okButton)

        wheel.setLayout(wheel_l)
        module.setLayout(module_l)

        layout.addWidget(wheel)
        layout.addWidget(module)
        layout.addLayout(panel)

        wrapper.setLayout(layout)

        wrapper.setWindowTitle("Edit actions")

        okButton.setProperty("wrapper", wrapper)
        okButton.setProperty("elem", elem)
        onShown.setProperty("wrapper", wrapper)

        return wrapper

    @pyqtSlot()
    def resizeCells(self):
        caller = self.sender()
        wrapper = caller.property("wrapper")

        wheel_l = (
            wrapper.findChild(QVBoxLayout).itemAt(0).widget().findChild(QGridLayout)
        )
        module_l = (
            wrapper.findChild(QVBoxLayout).itemAt(1).widget().findChild(QGridLayout)
        )

        title_width = max([l.cellRect(0, 1).width() for l in [wheel_l, module_l]])
        wheel_l.setColumnMinimumWidth(1, title_width)
        module_l.setColumnMinimumWidth(1, title_width)

    @pyqtSlot()
    def applyChanges(self):
        caller = self.sender()
        wrapper = caller.property("wrapper")
        elem = caller.property("elem")

        if wrapper is None or elem is None:
            self.logger.error("Could not get apply button propery")
            return

        actions = []

        for i, group in enumerate(
            [wrapper.findChild(QVBoxLayout).itemAt(x).widget() for x in [0, 1]]
        ):
            grid = group.findChild(QGridLayout)

            for j in range(1, grid.rowCount()):
                enabled_button = grid.itemAtPosition(j, 0)
                if enabled_button is None:
                    continue

                enabled = enabled_button.widget().isChecked()
                if not enabled:
                    continue

                onState = grid.itemAtPosition(j, 3).widget().currentText().lower()
                if i == 0:  # wheel
                    mode = "wheel"

                if i == 1:  # module
                    mode = "module"

                anyState = False
                if onState == "anywhere":
                    anyState = True

                actionName = grid.itemAtPosition(j, 0).widget().property("action_name")

                action = {
                    "action": actionName,
                    "mode": mode,
                    "checkState": not anyState,
                }

                if not anyState and not mode == onState:
                    action["onState"] = onState

                actions.append(action)

        # Saving
        exists, elem_bind = self.value_getter(
            "actionengine", "commandBind." + elem["device"]
        )
        if not exists:
            elem_bind = [{"command": elem["command"], "actions": []}]

        found = False
        for i, e in enumerate(elem_bind):
            if e.get("command") is not None and e["command"] == elem["command"]:
                found = True
                elem_bind[i]["actions"] = actions

        if not found:
            elem_bind.append({"command": elem["command"], "actions": actions})

        self.value_setter(
            module="actionengine", prop="commandBind." + elem["device"], value=elem_bind
        )


handlers = {"actions_picker": ActionPicker, "actions_list": ActionList}



import logging
import os

from pynput.keyboard import Controller, Key, KeyCode

from smartwheel.actions.baseaction import BaseAction


class Action(BaseAction):
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.keeb = Controller()
        self.media_keys = {
            "play_pause": Key.media_play_pause,
            "next": Key.media_next,
            "previous": Key.media_previous,
            "down": Key.media_volume_down,
            "up": Key.media_volume_up,
            "mute": Key.media_volume_mute,
        }

    def run(self, context):
        if context["type"] == "up":
            up = True
            down = False
        elif context["type"] == "down":
            up = False
            down = True
        else:
            up = True
            down = True

        if context["key_class"] == "special":
            if hasattr(Key, context["key"]):
                for _ in range(context["repeat"] + 1):
                    if down:
                        self.keeb.press(getattr(Key, context["key"]))
                    if up:
                        self.keeb.release(getattr(Key, context["key"]))
            else:
                self.logger.warning(
                    "Error: cannot parse Key.",
                    context["key"],
                    ". Please check pynput.Key class",
                    sep="",
                )
                self.logger.info("Key context:", context)
                return False

        elif context["key_class"] == "regular":
            for _ in range(context["repeat"] + 1):
                if down:
                    self.keeb.press(context["key"])
                if up:
                    self.keeb.release(context["key"])

        elif context["key_class"] == "vk":
            for _ in range(context["repeat"] + 1):
                if down:
                    self.keeb.press(KeyCode.from_vk(int(context["key"], 16) + 0x200))
                if up:
                    self.keeb.release(KeyCode.from_vk(int(context["key"], 16) + 0x200))

        elif context["key_class"] == "xdotool":
            for _ in range(context["repeat"] + 1):
                if down:
                    os.system("xdotool keydown " + context["key"])
                if up:
                    os.system("xdotool keyup " + context["key"])

        elif context["key_class"] == "media":
            key = self.media_keys.get(context["key"])
            if key is None:
                self.logger.warning("No such media key: " + context["key"])
                return False

            for _ in range(context["repeat"] + 1):
                if down:
                    self.keeb.press(key)
                if up:
                    self.keeb.release(key)
        else:
            return False
        return True



import logging
import math
import os

from PyQt6.QtCore import *
from PyQt6.QtGui import *

from smartwheel import config
from smartwheel.ui.base import BaseUIElem


class UIElem(BaseUIElem):
    def __init__(self, config_file, WConfig):
        super().__init__()
        self.config_file = config_file
        self.logger = logging.getLogger(__name__)
        self.loadConfig()
        self.conf.c = {**self.conf, **WConfig}
        self.icon_path = self.conf.get("icon_path", None)
        self.initGUI()
        self.initWidthAnimation()

    def loadConfig(self):
        self.conf = config.Config(self.config_file)
        self.conf.loadConfig()

    def checkOverflow(self, val):
        if val > 359:
            return val - 359
        elif val < 0:
            return 359 - val
        return val

    def processKey(self, event):
        if event["call"] == "keyAction1":
            self.mode = (self.mode + 1) % 3
            self.startWidthAnimation()
        elif event["call"] == "scrollUp" or event["call"] == "scrollDown":
            dlt = self.angle_delta if event["call"] == "scrollUp" else -self.angle_delta
            if self.mode == 0:
                self.hue_selection = self.checkOverflow(self.hue_selection + dlt)
            elif self.mode == 1:
                self.sat_selection = self.checkOverflow(self.sat_selection + dlt)
            elif self.mode == 2:
                self.bri_selection = self.checkOverflow(self.bri_selection + dlt)

            color = QColor.fromHsl(
                int(self.hue_selection),
                int(self.to_map(self.sat_selection)),
                int(self.to_map(self.bri_selection)),
            )
            r, g, b, _ = color.getRgbF()
            self.logger.debug(r, g, b)

            data = self.conf["internal"]["kritaAPI"]["class"].setColor(r, g, b)
            self.conf["internal"]["kritaServer"]["signals"]["send"].emit(data)
            # print(self.hue_selection, self.sat_selection, self.bri_selection)

    def initGUI(self):
        self.hue_selection = self.conf["HueStartAngle"]
        self.sat_selection = self.conf["SatStartAngle"]
        self.bri_selection = self.conf["BriStartAngle"]
        self.mode = 0
        self.angle_delta = self.to_map(1, 0, 359, 0, 255)

        self.delta = 360 // self.conf["selectionWheelEntries"]
        if self.conf["isHSLWidthFixed"]:
            self.hsl_width = self.conf["HSLCircleWidth"]
        else:
            self.hsl_width = self.conf["width"] // 8
        if self.conf["isHSLSelectionWidthFixed"]:
            self.hsl_select_width = self.conf["HSLSelectedCircleWidth"]
        else:
            self.hsl_select_width = self.conf["width"] // 6
        if self.conf["isHSLPaddingFixed"]:
            self.hsl_padding = self.conf["HSLPadding"]
        else:
            self.hsl_padding = self.conf["width"] // 16
        self.hsl_sw = self.hsl_select_width - self.hsl_width

        self.hsl_elem_width = [self.hsl_select_width, self.hsl_width, self.hsl_width]
        self.hsl_padding_width = [self.hsl_padding, self.hsl_padding, self.hsl_padding]

        self.color_window_width = (
            (self.conf["width"] * 3) / 4
            - self.hsl_width * 2
            - self.hsl_select_width
            - self.hsl_padding * 3
            - self.conf["colorWidgetPadding"]
        )

        # ---
        # self.kr_handler = KritaHandler()
        # self.kr_api = KritaAPI()
        # ---

    def _set_width(self, w):
        self.hsl_sw = w

    prop_width = pyqtProperty(int, fset=_set_width)

    def initWidthAnimation(self):
        self.is_anim_running = False
        self.width_anim_start = 0
        self.width_anim_end = self.hsl_select_width - self.hsl_width
        self.width_anim = QPropertyAnimation(self, b"prop_width")
        self.width_anim.setDuration(self.conf["widthAnimDuration"])
        self.width_anim.setStartValue(self.width_anim_start)
        self.width_anim.setEndValue(self.width_anim_end)

    def startWidthAnimation(self):
        self.is_anim_running = True
        # self.width_anim.stop()
        self.width_anim.start()

    def to_map(self, value, i_min=0, i_max=359, o_min=0, o_max=255):
        return o_min + (o_max - o_min) * ((value - i_min) / (i_max - i_min))

    def getX(self, a, w):
        return math.cos(math.radians(a)) * w / 2 + self.conf["cx"]

    def getY(self, a, h):
        return math.sin(math.radians(a)) * h / 2 + self.conf["cy"]

    def drawHueWheel(self, qp, width):
        width -= sum(self.hsl_elem_width[1:3]) + self.hsl_padding * 2
        if self.mode == 0:
            hsl_w = self.hsl_width + self.hsl_sw
        elif self.mode == 1:
            hsl_w = self.hsl_select_width - self.hsl_sw
        else:
            hsl_w = self.hsl_width
        self.hsl_elem_width[0] = hsl_w

        grad = QConicalGradient(
            QPointF(self.conf["cx"], self.conf["cy"]),
            359
            - self.hue_selection
            + self.conf["selectionAngle"]
            - 90
            - (self.delta // 4),
        )
        grad.setColorAt(0, QColor(255, 0, 0))
        grad.setColorAt(1.0 / 6, QColor(255, 255, 0))
        grad.setColorAt(2.0 / 6, QColor(0, 255, 0))
        grad.setColorAt(3.0 / 6, QColor(0, 255, 255))
        grad.setColorAt(4.0 / 6, QColor(0, 0, 255))
        grad.setColorAt(5.0 / 6, QColor(255, 0, 255))
        grad.setColorAt(1, QColor(255, 0, 0))

        qp.setPen(QPen(grad, 1))
        qp.setBrush(QBrush(grad))
        qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), width // 2, width // 2)
        self.drawPadding(qp, width - hsl_w)

        a = self.conf["selectionAngle"] - 90 + (self.delta // 4)
        self.drawColorPoint(
            qp,
            self.getX(a + 90.0, width - hsl_w / 2),
            self.getY(a + 90, width - hsl_w / 2),
            (hsl_w) / 4,
            QColor.fromHsl(int(self.hue_selection), 255, 128),
        )

    def drawSatWheel(self, qp, width):
        width -= self.hsl_elem_width[2] + self.hsl_padding
        if self.mode == 1:
            hsl_w = self.hsl_width + self.hsl_sw
        elif self.mode == 2:
            hsl_w = self.hsl_select_width - self.hsl_sw
        else:
            hsl_w = self.hsl_width
        self.hsl_elem_width[1] = hsl_w

        grad = QConicalGradient(
            QPointF(self.conf["cx"], self.conf["cy"]),
            359
            - self.sat_selection
            + self.conf["selectionAngle"]
            - 90
            - (self.delta // 4),
        )

        grad.setColorAt(0, QColor.fromHsl(int(self.hue_selection), 0, 128))
        grad.setColorAt(1, QColor.fromHsl(int(self.hue_selection), 255, 128))

        qp.setPen(QPen(grad, 1))
        qp.setBrush(QBrush(grad))
        qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), width // 2, width // 2)
        self.drawPadding(qp, width - hsl_w)

        a = self.conf["selectionAngle"] - 90 + (self.delta // 4)
        self.drawColorPoint(
            qp,
            self.getX(a + 90.0, width - hsl_w / 2),
            self.getY(a + 90, width - hsl_w / 2),
            (hsl_w) / 4,
            QColor.fromHsl(
                int(self.hue_selection), int(self.to_map(self.sat_selection)), 128
            ),
        )

    def drawBriWheel(self, qp, width):
        hsl_w = self.hsl_width
        if self.mode == 2:
            hsl_w = self.hsl_width + self.hsl_sw
        elif self.mode == 0:
            hsl_w = self.hsl_select_width - self.hsl_sw
        self.hsl_elem_width[2] = hsl_w

        grad = QConicalGradient(
            QPointF(self.conf["cx"], self.conf["cy"]),
            359
            - self.bri_selection
            + self.conf["selectionAngle"]
            - 90
            - (self.delta // 4),
        )

        grad.setColorAt(0, QColor(0, 0, 0))
        grad.setColorAt(
            0.5,
            QColor.fromHsl(
                int(self.hue_selection), int(self.to_map(self.sat_selection)), 128
            ),
        )
        grad.setColorAt(1, QColor(255, 255, 255))

        qp.setPen(QPen(grad, 1))
        qp.setBrush(QBrush(grad))
        qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), width // 2, width // 2)
        self.drawPadding(qp, width - hsl_w)

        a = self.conf["selectionAngle"] - 90 + (self.delta // 4)

        self.drawColorPoint(
            qp,
            self.getX(a + 90.0, width - hsl_w / 2),
            self.getY(a + 90, width - hsl_w / 2),
            (hsl_w) / 4,
            QColor.fromHsl(
                int(self.hue_selection),
                int(self.to_map(self.sat_selection)),
                int(self.to_map(self.bri_selection)),
            ),
        )

    def drawColorPoint(self, qp, cx, cy, radius, color):
        if qGray(color.rgb()) >= self.conf["colorDotLightnessThreshold"]:
            qp.setPen(
                QPen(
                    QColor(self.conf["colorDotDarkStrokeColor"]),
                    self.conf["colorDotStrokeWidth"],
                )
            )
        else:
            qp.setPen(
                QPen(
                    QColor(self.conf["colorDotLightStrokeColor"]),
                    self.conf["colorDotStrokeWidth"],
                )
            )

        radius -= 0

        qp.setBrush(QBrush(color))
        qp.drawEllipse(QPointF(cx, cy), radius, radius)

    def drawPadding(self, qp, width):
        qp.setPen(QPen())
        qp.setBrush(QBrush(QColor(self.conf["bgWheelColor"])))
        qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), width // 2, width // 2)

    def drawColorWindow(self, qp, color, width, offset):
        qp.setPen(
            QPen(
                QColor(self.conf["colorDotLightStrokeColor"]),
                self.conf["colorDotStrokeWidth"],
            )
        )

        width = min(width, self.conf["colorWidgetMaxWidth"] / 2)

        qp.setBrush(QBrush(color))
        if self.conf["colorWidgetShape"] == "circle":
            qp.drawEllipse(QPoint(self.conf["cx"], self.conf["cy"]), width, width)
        elif self.conf["colorWidgetShape"] == "square":
            qp.drawRect(
                QRect(
                    self.conf["cx"] - width,
                    self.conf["cy"] - width,
                    width * 2,
                    width * 2,
                )
            )
        elif self.conf["colorWidgetShape"] == "hybrid":
            qp.drawRoundedRect(
                QRectF(
                    self.conf["cx"] - width,
                    self.conf["cy"] - width,
                    width * 2,
                    width * 2,
                ),
                self.conf["wheelWidth"] - offset,
                self.conf["wheelWidth"] - offset,
                Qt.SizeMode.RelativeSize,
            )

    def draw(self, qp, offset=None):
        if self.width_anim.state() == QAbstractAnimation.State.Stopped:
            self.is_anim_running = False
        self.drawBriWheel(qp, (self.conf["width"] * 3) // 4 + offset)
        self.drawSatWheel(qp, (self.conf["width"] * 3) // 4 + offset)
        self.drawHueWheel(qp, (self.conf["width"] * 3) // 4 + offset)

        self.drawColorWindow(
            qp,
            QColor.fromHsl(
                int(self.hue_selection),
                int(self.to_map(self.sat_selection)),
                int(self.to_map(self.bri_selection)),
            ),
            (self.color_window_width + offset) / 2,
            offset,
        )



import io
import os
import random
import string

import matplotlib.pyplot as plt
import numpy as np
from PyQt6.QtGui import QPixmap

from smartwheel import common, tools
from smartwheel.backgrounds.base import Background


class ContourBackground(Background):
    def __init__(self, common_config, conf, canvas):
        super(ContourBackground, self).__init__(common_config, conf, canvas)
        tools.merge_dicts(
            self.conf, self.common_config(), include_only=["wheelTextureColor"]
        )
        self.setTexture(self.loadPixmap())

        self.conf.updateFunc = self.updatePixmap
        # self.conf.updated.connect(self.updatePixmap)

    def genContour(self):
        if self.conf["randomSeed"]:
            seed = random.randint(0, 100000)
            self.conf["seed"] = seed

        random.seed(self.conf["seed"])

        plt.rcParams["figure.facecolor"] = self.conf["backgroundColor"]
        plt.figure(figsize=(1, 1), dpi=self.common_config()["width"] + 10)
        plt.axes([0, 0, 1, 1])

        def perlin(x, y, seed=0):
            # permutation table
            np.random.seed(seed)
            p = np.arange(256, dtype=int)
            np.random.shuffle(p)
            p = np.stack([p, p]).flatten()
            # coordinates of the top-left
            xi, yi = x.astype(int), y.astype(int)
            # internal coordinates
            xf, yf = x - xi, y - yi
            # fade factors
            u, v = fade(xf), fade(yf)
            # noise components
            n00 = gradient(p[p[xi] + yi], xf, yf)
            n01 = gradient(p[p[xi] + yi + 1], xf, yf - 1)
            n11 = gradient(p[p[xi + 1] + yi + 1], xf - 1, yf - 1)
            n10 = gradient(p[p[xi + 1] + yi], xf - 1, yf)
            # combine noises
            x1 = lerp(n00, n10, u)
            x2 = lerp(n01, n11, u)  # FIX1: I was using n10 instead of n01
            return lerp(x1, x2, v)  # FIX2: I also had to reverse x1 and x2 here

        def lerp(a, b, x):
            "linear interpolation"
            return a + x * (b - a)

        def fade(t):
            "6t^5 - 15t^4 + 10t^3"
            return 6 * t**5 - 15 * t**4 + 10 * t**3

        def gradient(h, x, y):
            "grad converts h to the right gradient vector and return the dot product with (x,y)"
            vectors = np.array([[0, 1], [0, -1], [1, 0], [-1, 0]])
            g = vectors[h % 4]
            return g[:, :, 0] * x + g[:, :, 1] * y

        sign = lambda: (-1) ** int(random.randint(1, 2))

        def getXY(x, y):
            a = random.randint(0, 2)

            if a == 0:
                return sign() * x + sign() * y + random.randint(-20, 20)
            if a == 1:
                return sign() * x + random.randint(-20, 20)
            if a == 2:
                return sign() * y + random.randint(-20, 20)

        def sine_family(x, y):
            return sign() * np.sin(getXY(x, y)) + sign() * np.cos(getXY(x, y))

        def perlin_gen(x, y):
            return perlin(x, y, seed=self.conf["seed"])

        # Generating data
        x = np.linspace(0, 5, 50)
        y = np.linspace(0, 5, 50)
        x, y = np.meshgrid(x, y)

        if self.conf["generator"] == "Perlin":
            z = perlin_gen(x, y)
        else:
            z = sine_family(x, y)

        # Plotting the contour plot
        plt.contour(x, y, z, colors=self.conf["wheelTextureColor"])

        # Adding details to the plot
        plt.title("sin(x) + cos(10+y*x)")
        plt.xlabel("x-axis")
        plt.ylabel("y-axis")
        plt.axis("off")
        plt.box("off")

        img_buf = io.BytesIO()
        plt.savefig(img_buf, format="png")
        img_buf.seek(0)

        plt.close()

        if self.conf["useCache"]:
            filepath = common.cache_manager.save(
                "background_contour", "contour1.png", self.conf
            )
            with open(filepath, "wb") as f:
                f.write(img_buf.getvalue())

        img_buf.seek(0)

        return self.contourToPixmap(img_buf)

    def checkCache(self, conf):
        params = ["backgroundColor", "wheelTextureColor", "seed"]

        for p in params:
            if not self.conf.get(p) == conf.get(p):
                return False
        return True

    def updatePixmap(self):
        pix = self.genContour()
        self.setTexture(pix)

    def loadPixmap(self):
        if self.conf["useCache"] and not self.conf["randomSeed"]:
            ok, filepath, conf = common.cache_manager.load(
                "background_contour", "contour1.png"
            )
            if ok and self.checkCache(conf):
                pix = QPixmap()
                pix.load(filepath)
                return pix

        pix = self.genContour()

        return pix

    def contourToPixmap(self, img_buf):
        bytes = img_buf.read()
        pix = QPixmap()
        pix.loadFromData(bytes, format="png")
        return pix


brushes = {"contour": ContourBackground}



import logging

from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (
    QCheckBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from smartwheel.settings_handlers.base import BaseHandler


class ModulesLoader(BaseHandler):
    """
    Common modules picker
    """

    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(ModulesLoader, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        """
        Initialize modules picker, should not be called directly from settings.py

        Parameters
        ==========
        elem
            Dict containing modules {"modules": [{"name": "internalName", "title": ..., "description": ...,}, ...], "modulesLoad": [0, 1, ...]}
        """
        layout = QGridLayout()

        font = QFont()
        font.setBold(True)

        labels = [QLabel(x) for x in ["Enabled", "Title", "Description", "Options"]]

        for i in range(len(labels)):
            labels[i].setFont(font)
            layout.addWidget(labels[i], 0, i, Qt.AlignmentFlag.AlignLeft)

        # layout.setColumnStretch(10, 0)

        for i, mod in enumerate(elem["modules"]):
            check = QCheckBox()
            options = QPushButton("...")

            check.setProperty("name", mod["name"])

            if i in elem["modulesLoad"]:
                check.setChecked(True)

            for j, s in enumerate(["title", "description"]):
                label = QLabel(mod[s])

                layout.addWidget(label, i + 1, j + 1, Qt.AlignmentFlag.AlignLeft)

            layout.addWidget(check, i + 1, 0, Qt.AlignmentFlag.AlignLeft)
            layout.addWidget(options, i + 1, 3, Qt.AlignmentFlag.AlignLeft)

            for j in range(1, 4):
                if layout.itemAtPosition(i + 1, j) is not None:
                    check.clicked.connect(
                        layout.itemAtPosition(i + 1, j).widget().setEnabled
                    )
                    if not check.isChecked():
                        layout.itemAtPosition(i + 1, j).widget().setDisabled(True)

        vbox = QVBoxLayout()
        vbox.addLayout(layout)

        spacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        vbox.addSpacerItem(spacer)

        wrapper = QWidget()
        wrapper.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        wrapper.setLayout(vbox)

        return wrapper


class SerialLoader(BaseHandler):
    """
    Serial modules picker
    """

    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(SerialLoader, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        """
        Load serial modules picker

        Parameters
        ==========
        elem
            Element from config
        """
        ok, modules = self.value_getter("canvas", "serialModules")
        if not ok:
            self.logger.error("Could not get serial modules")
            return None

        ok, modulesLoad = self.value_getter("canvas", "serialModulesLoad")
        if not ok:
            self.logger.error("Could not get loaded serial modules")
            return None

        elem["modules"] = modules
        elem["modulesLoad"] = modulesLoad

        picker = self.parent_obj().handlers["modules"].initElem(elem)
        if picker is None:
            self.logger.error("Could not initialize modules picker")
            return None

        layout = picker.findChild(QVBoxLayout).findChild(QGridLayout)

        for i in range(1, layout.rowCount()):
            if layout.itemAtPosition(i, 0) is None:
                break

            checked = layout.itemAtPosition(i, 0).widget()
            checked.clicked.connect(self.setEnabled)

        return picker

    @pyqtSlot(bool)
    def setEnabled(self, enabled):
        caller = self.sender()

        name = caller.property("name")

        ok, modules = self.value_getter("canvas", "serialModules")
        if not ok:
            self.logger.error("Could not get serial modules")
            return

        ok, modulesLoad = self.value_getter("canvas", "serialModulesLoad")
        if not ok:
            self.logger.error("Could not get loaded serial modules")
            return

        index = None
        for i in range(len(modules)):
            if modules[i]["name"] == name:
                index = i
                break

        if index is None:
            self.logger.error("Could not get module by name")
            return

        if enabled:
            m_set = set(modulesLoad)
            m_set.add(index)
            modulesLoad = list(m_set)
        else:
            modulesLoad.remove(index)

        self.value_setter(module="canvas", prop="serialModulesLoad", value=modulesLoad)


handlers = {"modules": ModulesLoader, "serial": SerialLoader}



import importlib
import json
import logging
import os
import queue
import time
import weakref

from PyQt6.QtCore import *
from PyQt6.QtGui import *

from smartwheel import common, config, gui_tools
from smartwheel.actionengine import ActionEngine
from smartwheel.tools import merge_dicts


class MList(list):
    """Overloaded list class to store modules. Used to work with weakref"""

    pass


class MDict(dict):
    """Overloaded dict class that represents a module. Uses weakref"""

    pass


class RootCanvas(QObject):
    """Main canvas class, manages wheel modules"""

    def __init__(self, WConfig, config_dir, update_func):
        super(RootCanvas, self).__init__()
        self.common_config = None
        self.config_dir = config_dir
        self.conf = WConfig
        self.update_func = update_func
        self.logger = logging.getLogger(__name__)
        self.loadCommonConf()
        self.processCommonConfig()
        # TODO add try catch everywhere
        self.initCacheDir()
        self.loadInternalModules()
        self.loadBrushes()
        self.startInternalModules()
        self.pool = QThreadPool.globalInstance()
        self.threads = []
        self.wheel_modules = self.loadSections(self.conf["wheelModules"])
        self.cur_wheel_modules = self.wheel_modules
        if self.wheel_modules == 1:
            raise KeyError
        self.loadModules()
        self.tick = 0
        self.loadActionEngine()

        self.exec_time = 0.01
        self.exec_window = 0
        self.exec_times = queue.Queue()

        self.startThreads()

    def loadCommonConf(self):
        self.common_config = config.Config(
            os.path.join(self.config_dir, self.conf["commonConfig"]), self.logger
        )
        if not self.common_config.loadConfig():
            self.logger.error("Could not find common config file. Exiting..")
            os._exit(1)

        self.common_config["iconsFolder"] = os.path.join(
            self.conf["basedir"], self.common_config["iconsFolder"]
        )
        self.updateIconCache()

    def loadModules(self):
        """
        Read module classes from `modules` config
        """
        for i in self.conf["modulesLoad"]:
            self.conf["modules"][i]["class"] = self.importModule(
                self.conf["modules"][i]
            )

    def importModule(self, meta):
        """
        Import a single module class

        Parameters
        ==========
        meta
            Configuration of the module
        """
        mod = importlib.import_module("smartwheel." + meta["name"])
        ui = mod.UIElem(
            os.path.join(self.config_dir, meta["config"]),
            self.conf,
            self.wheel_modules,
            self.update_func,
            weakref.ref(self),
        )
        return ui

    def loadSections(self, modules_list, parent_mod=None):
        """
        Configure already loaded modules or recursively load them in folders

        Parameters
        ==========
        modules_list
            `wheelModules` option from config
        parent_mod
            Folder module (None if it's a root)
        """
        mods = MList()
        for i in modules_list:
            mod_dict = next(
                (m for m in self.conf["modules"] if m["name"] == i["name"]), None
            )
            if i["name"] == "ui.folder":
                mod = MDict(i)
                submod = self.loadSections(mod["modules"], mods)
                if submod == 1:
                    return 1
                mod["modules"] = submod
                # mod["parent"] = weakref.ref(mods)
            elif mod_dict is None:
                continue
            else:
                mod = MDict(mod_dict)
            if parent_mod is not None:
                mod["parent"] = weakref.ref(parent_mod)
            # elif mod is None:
            #    print("Error importing module ", mod["name"], ": No such module. Aborting...",sep="")
            #    return 1
            mod_class = self.loadWheelModule(mod)
            if i["name"] == "ui.folder":
                mod_class["class"].wrapper_pointer = weakref.ref(mod_class)
            mods.append(mod_class)
        return mods

    def initCacheDir(self):
        common.cache_manager.initManager(self.conf)

    def loadBrushes(self):
        b_config = os.path.join(
            self.conf["basedir"], self.conf["brushes_dir"], "config.json"
        )
        if not os.path.exists(b_config):
            self.logger.error("Missing " + b_config + " file")
            os._exit(1)

        with open(b_config, "r") as f:
            self.brushes_conf = json.load(f)

        self.brushes = {}
        self.conf["brush_configs"] = {}

        for mod_name in self.brushes_conf["brushes_modules"]:
            brush = importlib.import_module(
                "smartwheel." + self.conf["brushes_dir"] + "." + mod_name
            )
            b_dict = brush.brushes
            for k in b_dict:
                if self.brushes.get(k) is not None:
                    self.logger.warning("Brush " + k + " is defined twice. Overriding")

                self.conf["brush_configs"][k] = config.Config(
                    os.path.join(
                        self.conf["basedir"],
                        self.conf["brushes_dir"],
                        self.brushes_conf["brushes_config_dir"],
                        self.brushes_conf["brushes_config"][k],
                    )
                )
                self.conf["brush_configs"][k].loadConfig()

                self.brushes[k] = b_dict[k](
                    weakref.ref(self.conf),
                    self.conf["brush_configs"][k],
                    weakref.ref(self),
                )

        self.conf["brushesTypes"] = list(self.brushes.keys())

    def processCommonConfig(self):
        """
        Add some properties to the common config and merge it with self.conf
        """
        if self.common_config["isWheelWidthFixed"]:
            self.common_config["wheelWidth"] = self.common_config["fixedWheelWidth"]
        else:
            self.common_config["wheelWidth"] = self.conf["width"] // 4
        self.common_config["configDir"] = self.config_dir

        # Merging configs in order to make settings editable on-the-fly
        merge_dicts(self.conf, self.common_config)

    def loadWheelModule(self, module):
        mod = importlib.import_module("smartwheel." + module["name"])
        if module.get("config", None) is not None:
            ui = mod.UIElem(os.path.join(self.config_dir, module["config"]), self.conf)

            if hasattr(ui, "updateSignal"):
                ui.updateSignal.connect(self.updateCanvas)

            if hasattr(ui, "thread"):
                self.threads.append(ui.thread)
        else:
            ui = mod.UIElem("", module)
        module["class"] = ui
        # self.pixmap = QImage(self.module["class"].icon_path)
        return module

    @pyqtSlot()
    def updateCanvas(self):
        """
        Call update event to refresh canvas
        """
        self.update_func()

    def startThreads(self):
        for thread in self.threads:
            self.pool.start(thread)

    @pyqtSlot()
    def killThreads(self):
        for thread in self.threads:
            if hasattr(thread, "shutdown"):
                thread.shutdown = True

        self.pool.waitForDone(100)

    def reloadWheelModules(self, is_up, caller=None):
        if is_up:
            if caller["parent"] is not None:
                self.cur_wheel_modules = caller["parent"]()
                # return caller["parent"]()
            # return None
        else:
            self.cur_wheel_modules = caller["modules"]
            # return caller["modules"]
        self.conf["modules"][0]["class"].reloadModules(self.cur_wheel_modules)

    def loadInternalModules(self):
        self.conf["internal"] = {}
        for i in self.conf["internalModulesLoad"]:
            mod = self.conf["internalModules"][i]
            if mod.get("config") is not None:
                cnf = os.path.join(self.config_dir, mod["config"])
            else:
                cnf = None
            mod_class = importlib.import_module("smartwheel." + mod["name"]).Internal(
                self.conf, cnf
            )
            self.conf["internal"][mod_class.name] = {
                "class": mod_class,
                "signals": mod_class.getSignals(),
            }

    def startInternalModules(self):
        for i in self.conf["internal"]:
            self.conf["internal"][i]["class"].start()

    def getCurModList(self):
        """
        Get current wheel modules
        """
        return self.cur_wheel_modules

    def getWheelModule(self):
        i = self.conf["modules"][0]["class"].getCurModule()
        if i >= len(self.cur_wheel_modules):
            return None
        return self.cur_wheel_modules[i]

    def loadActionEngine(self):
        self.ae = ActionEngine(
            self.wheel_modules,
            os.path.join(self.config_dir, self.conf["actionEngineConfig"]),
            self.conf,
        )
        self.ae.current_module_list_getter = self.getCurModList
        self.ae.current_module_getter = self.conf["modules"][0]["class"].getCurModule
        self.ae.canvas = weakref.ref(self)
        self.ae.wheel = weakref.ref(self.conf["modules"][0]["class"])

    def updateIconCache(self):
        """
        Will only be called if icon color has been changed
        """
        res1 = gui_tools.icon_managers["wheel"].setIconColor(
            self.common_config["wheelIconColor"]
        )
        return (
            gui_tools.icon_managers["sections"].setIconColor(
                self.common_config["sectionsIconColor"]
            )
            or res1
        )

    def calculateSmoothFPS(self, new_time):
        """
        Calculate the average render time out of n frames

        Parameters
        ==========
        new_time
            The most recent frame render time
        """
        if self.exec_window == 0:
            self.exec_times.put(new_time)
            self.exec_time = new_time
            self.exec_window += 1
            return

        if self.exec_window < self.conf["fpsFramesSmooth"]:
            self.exec_window += 1
            self.exec_times.put(new_time)
            self.exec_time = (
                (self.exec_window - 1) * self.exec_time + new_time
            ) / self.exec_window
            return

        last_time = self.exec_times.get()
        self.exec_times.put(new_time)

        self.exec_time = (
            self.exec_window * self.exec_time - last_time + new_time
        ) / self.exec_window

    def draw(self, qp):
        """
        Main draw function

        Parameters
        ----------
        qp
            QPainter object
        """
        # for i in self.conf["modulesLoad"]:
        #    self.conf["modules"][i]["class"].draw(qp)
        self.conf["modules"][0]["class"].draw(qp)  # render wheel
        m = self.getWheelModule()
        if (
            self.conf["modules"][0]["class"].is_anim_running
            or (
                m is not None
                and hasattr(m["class"], "is_anim_running")
                and m["class"].is_anim_running
            )
            or self.updateIconCache()
        ):
            if self.conf["stabilizeFPS"]:
                sleep_time = max(1 / self.conf["fps"] - self.exec_time, 0)
            else:
                sleep_time = 1 / self.conf["fps"]

            time.sleep(sleep_time)

            start_time = time.time()

            self.update_func()  # TODO add another update event

            e_time = (time.time() - start_time) * 1000

            if self.conf["stabilizeFPS"]:
                if self.conf["logFPS"]:
                    self.logger.info(
                        "FPS(AVG): "
                        + str(round(1 / max(sleep_time + self.exec_time, 0.0000001), 1))
                    )
                self.calculateSmoothFPS(e_time)

            else:
                if self.conf["logFPS"]:
                    self.logger.info("FPS: " + str(round(1 / (sleep_time + e_time), 1)))



import json

from PyQt6.QtCore import pyqtSignal, pyqtSlot

from smartwheel.ui.internal.baseinternal import BaseInternal


class Internal(BaseInternal):
    def __init__(self, WConfig, config_file):
        super().__init__()
        self.name = "kritaAPI"

    def setColor(self, r, g, b):
        call = {"action": "set_color"}
        call["properties"] = {"rgb": [r, g, b]}
        return json.dumps(call)



from PyQt6.QtCore import QThread


class BaseInternal(QThread):
    def __init__(self):
        super().__init__()
        self.name = "baseInternal"
        self.signals = {}

    def __del__(self):
        pass

    def getSignals(self):
        return self.signals

    def run(self):
        pass



import logging

from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtGui import QColor, QIcon, QPixmap
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QLineEdit,
    QPushButton,
    QSpinBox,
)
from swcolorpicker import getColor, rgb2hex, useAlpha

from smartwheel.settings_handlers.base import BaseHandler


class IntHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(IntHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        wid = QSpinBox()
        if elem.get("min") is not None:
            wid.setMinimum(elem["min"])
        if elem.get("max") is not None:
            wid.setMaximum(elem["max"])

        ok, value = self.value_getter(elem["module"], elem["prop"], elem.get("index"))
        if ok:
            wid.setValue(value)
        else:
            self.logger.warning("Could not get value for " + elem["name"])

        wid.valueChanged.connect(self.setInt)

        return wid

    @pyqtSlot(int)
    def setInt(self, value):
        caller = self.sender()
        self.value_setter(caller, value)

    def fetchValue(self, wid):
        return wid.value()

    def updateValue(self, wid, value):
        wid.setValue(value)
        return True


class FloatHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(FloatHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        wid = QDoubleSpinBox()

        if elem.get("step") is not None:
            wid.setSingleStep(elem["step"])

        if elem.get("min") is not None:
            wid.setMinimum(elem["min"])
        if elem.get("max") is not None:
            wid.setMaximum(elem["max"])

        ok, value = self.value_getter(elem["module"], elem["prop"], elem.get("index"))
        if ok:
            wid.setValue(value)
        else:
            self.logger.warning("Could not get value for " + elem["name"])

        wid.valueChanged.connect(self.setFloat)

        return wid

    @pyqtSlot(float)
    def setFloat(self, value):
        caller = self.sender()
        self.value_setter(caller, value)

    def fetchValue(self, wid):
        return wid.value()

    def updateValue(self, wid, value):
        wid.setValue(value)
        return True


class StringHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(StringHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        wid = QLineEdit()

        ok, value = self.value_getter(elem["module"], elem["prop"])
        if ok:
            wid.setText(value)
        else:
            self.logger.warning("Could not get value for " + elem["name"])

        wid.textChanged.connect(self.setStr)

        return wid

    @pyqtSlot(str)
    def setStr(self, value):
        caller = self.sender()
        self.value_setter(caller, value)

    def fetchValue(self, wid):
        return wid.text()

    def updateValue(self, wid, value):
        wid.setText(value)
        return True


class ColorHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(ColorHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def setIcon(self, wid, color):
        pix = QPixmap(100, 100)
        pix.fill(color)
        icon = QIcon(pix)
        wid.setIcon(icon)

    def initElem(self, elem):
        wid = QPushButton()

        wid.setStyleSheet("QPushButton { text-align: left; }")

        wid.setProperty("module", elem["module"])
        wid.setProperty("prop", elem["prop"])

        ok, value = self.value_getter(elem["module"], elem["prop"])
        if ok:
            wid.setText(value)
        else:
            self.logger.warning("Could not get value for " + elem["name"])

        self.setIcon(wid, QColor(value))

        wid.clicked.connect(self.setColor)

        return wid

    @pyqtSlot()
    def setColor(self):
        caller = self.sender()
        color = QColor(caller.text()).getRgb()
        if len(color) == 4:
            useAlpha(True)

        new_color = "#" + rgb2hex(getColor(color))
        if not color == new_color:
            self.value_setter(caller, new_color)
            self.setIcon(caller, QColor(new_color))
            caller.setText(new_color)

    def fetchValue(self, wid):
        return wid.text()

    def updateValue(self, wid, value):
        module = wid.property("module")
        prop = wid.property("prop")

        if module is None or prop is None:
            self.logger.warning("Could not get color picker properties")
            return False

        wid.setText(value)
        self.setIcon(wid, QColor(value))
        self.value_setter(value=value, module=module, prop=prop)
        return True


class BoolHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(BoolHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        wid = QCheckBox()

        ok, value = self.value_getter(elem["module"], elem["prop"])
        if ok:
            wid.setChecked(value)
        else:
            self.logger.warning("Could not get value for " + elem["name"])

        wid.clicked.connect(self.setBool)

        return wid

    @pyqtSlot(bool)
    def setBool(self, value):
        caller = self.sender()
        self.value_setter(caller, value)

    def fetchValue(self, wid):
        return wid.isChecked()

    def updateValue(self, wid, value):
        wid.setChecked(value)
        return True


class ComboHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(ComboHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        wid = QComboBox()

        if type(elem["options"]) == list:
            wid.insertItems(0, elem["options"])
        else:
            # Need to get options programatically
            ok, ops = self.value_getter(
                module=elem["options"]["module"],
                prop=elem["options"]["prop"],
                index=elem["options"].get("index"),
            )
            if not ok:
                self.logger.error(
                    "Could not get combo items from "
                    + elem["options"]["module"]
                    + "."
                    + elem["options"]["prop"]
                )
                return None
            wid.insertItems(0, ops)

        ok, value = self.value_getter(elem["module"], elem["prop"])
        if ok:
            wid.setCurrentText(value)
        else:
            self.logger.warning("Could not get value for " + elem["name"])

        wid.currentTextChanged.connect(self.setStr)

        return wid

    @pyqtSlot(str)
    def setStr(self, value):
        caller = self.sender()
        self.value_setter(caller, value)

    def fetchValue(self, wid):
        return wid.currentText()

    def updateValue(self, wid, value):
        wid.setCurrentText(value)
        return True

    def linkElem(self, elem, registriesName):
        elem.setProperty("registriesName", registriesName)
        elem.currentTextChanged.connect(self.parent_obj().showLinkedWidgets)
        return True


handlers = {
    "int": IntHandler,
    "float": FloatHandler,
    "string": StringHandler,
    "color": ColorHandler,
    "bool": BoolHandler,
    "combo": ComboHandler,
}



import logging

from pynput import keyboard
from PyQt6.QtCore import *

from smartwheel import config
from smartwheel.serialpipe.base import ClickButton, ConnPipe, PRButton, Rotary


class SConn(ConnPipe):
    """Background keyboard listener"""

    def __init__(self, config_file, call_signal):
        """
        Initialize keyboardpipe

        Parameters
        ----------
        config_file
            Configuration file
        call_signal
            Actionegnine signal to call
        """
        super().__init__()
        self.keys = None
        self.conf = None
        self.logger = logging.getLogger(__name__)
        self.config_file = config_file
        self.call = call_signal
        self.loadConfig()
        self.loadKeys()

    def loadConfig(self):
        self.conf = config.Config(self.config_file, self.logger)
        self.conf.loadConfig()

    def loadKeys(self):
        """
        Initialize dicts of keys to quickly find them
        """
        self.keys = {
            "keyboards": {},
            "prbuttons": {},
            "clickbuttons": {},
            "encoders": {},
        }
        self.prbuttons = {}
        self.clickbuttons = {}
        self.encoders = {}

        for kbs in self.conf["keyboards"]:
            for k in kbs["keys"]:
                self.keys["keyboards"][k["string"]] = (kbs, k)

        for btn in self.conf["prbuttons"]:
            self.keys["prbuttons"][btn["key"]] = btn
            self.prbuttons[btn["name"]] = PRButton(btn["threshold"])
            self.prbuttons[btn["name"]].setupCallbacks(
                [self.call] * 3,
                [(btn, {"string": x}) for x in ["press", "click", "doubleclick"]],
            )

        for btn in self.conf["clickbuttons"]:
            self.keys["clickbuttons"][btn["key"]] = btn
            self.clickbuttons[btn["name"]] = ClickButton(btn["threshold"])
            self.clickbuttons[btn["name"]].setupCallbacks(
                [self.call] * 2,
                [(btn, {"string": x}) for x in ["click", "doubleclick"]],
            )

        for enc in self.conf["encoders"]:
            self.keys["encoders"][enc["keyUp"]] = (enc, True)  # up
            self.keys["encoders"][enc["keyDown"]] = (enc, False)  # down

            if (
                enc["linkedButton"] is not None
                and self.prbuttons.get(enc["linkedButton"]) is None
            ):
                self.logger.error(
                    "Could not link "
                    + enc["name"]
                    + " with non-existent prbutton "
                    + enc["linkedButton"]
                )

            self.encoders[enc["name"]] = Rotary(self.prbuttons[enc["linkedButton"]])
            self.encoders[enc["name"]].setupCallbacks(
                [self.call] * 6,
                [
                    (enc, {"string": x})
                    for x in [
                        "up",
                        "down",
                        "click up",
                        "click down",
                        "double up",
                        "double down",
                    ]
                ],
            )

    def on_press(self, key):
        """
        Pynput on press event

        Parameters
        ----------
        key
            Key object
        """
        k = str(key).strip("'")

        if self.keys["keyboards"].get(k) is not None:
            kb, btn = self.keys["keyboards"][k]
            self.call.emit((kb, btn))

        if self.keys["prbuttons"].get(k) is not None:
            btn = self.keys["prbuttons"][k]
            self.prbuttons[btn["name"]].execPress(True)

    def on_release(self, key):
        """
        Pynput on release event

        Parameters
        ----------
        key
            Key object
        """
        k = str(key).strip("'")

        if self.keys["prbuttons"].get(k) is not None:
            btn = self.keys["prbuttons"][k]
            self.prbuttons[btn["name"]].execPress(False)

        if self.keys["clickbuttons"].get(k) is not None:
            btn = self.keys["clickbuttons"][k]
            self.clickbuttons[btn["name"]].execPress()

        if self.keys["encoders"].get(k) is not None:
            enc, up = self.keys["encoders"][k]
            self.encoders[enc["name"]].execRotate(up)  # TODO add audio keys workaround

    def run(self):
        """
        The main listener loop
        """
        with keyboard.Listener(
            on_press=self.on_press, on_release=self.on_release
        ) as listener:
            listener.join()



def merge_dicts(dest, src, include_only=None):
    """
    Merge the contents of 2 dictionaries/configs while preserving references

    Should be used instead of `dest = {**dest, **src}`

    Note: references in dict objects are not preserved, please use Config object instead

    Parameters
    ==========
    dest
        Destination dict/config object
    src
        Source dict/config object
    include_only
        (Optional) Merge only this values from src
    """

    if type(dest) == dict:
        dest.update(src)
    else:
        dest.update(src, include_only=include_only)



from PyQt6.QtGui import QBrush


class Background(QBrush):
    """
    Background object, inherited from QBrush. Please see Qt6 QBrush documentation

    Must contain brushes dict at the end of the file
    """

    def __init__(self, common_config=None, config=None, canvas=None, *args, **kwargs):
        super(Background, self).__init__(*args, **kwargs)
        self.common_config = common_config
        self.conf = config
        self.canvas = canvas


# Needs to be at the end of the file
# brushes = {}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Smartwheel Core"
copyright = "2022, enaix"
author = "enaix"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import os
import sys

if not os.path.exists("./_static"):
    os.mkdir("./_static")

sys.path.insert(0, os.path.abspath("../"))

extensions = ["sphinx.ext.autodoc", "numpydoc"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autoclass_content = "both"  # enable __self__

autodoc_mock_imports = ["pynput"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
import importlib
import json
import logging
import os
import sys
import weakref

import qdarktheme
from PyQt6 import QtCore
from PyQt6.QtCore import QEvent, pyqtSlot
from PyQt6.QtGui import *
from PyQt6.QtWidgets import (
    QApplication,
    QDockWidget,
    QGraphicsBlurEffect,
    QMainWindow,
    QPushButton,
    QWidget,
)

from smartwheel import config
from smartwheel.canvas import RootCanvas
from smartwheel.settings import SettingsWindow


class WConfig(config.Config):
    def __init__(self, config_file, launch_config):
        super(WConfig, self).__init__(config_file, varsWhitelist=["serialModulesLoad"])

        if not self.loadConfig():
            print("Fatal: error loading main config file. Exiting..")
            os.exit(1)

        self.launch_config = launch_config

        self.processConfig()

    def processConfig(self):
        self.c_canvas = config.Config(config_dict=self.c["canvas"])
        c_geometry = self.c["window"]["geometry"]
        self.c_canvas["width"] = c_geometry[2] - self.c["window"]["padding"] * 2
        self.c_canvas["height"] = c_geometry[3] - self.c["window"]["padding"] * 2
        self.c_canvas["cx"] = c_geometry[2] // 2
        self.c_canvas["cy"] = c_geometry[3] // 2
        self.c_canvas["corner_x"] = self.c["window"]["padding"]
        self.c_canvas["corner_y"] = self.c["window"]["padding"]


class RootWindow(QMainWindow):
    def __init__(self, conf):
        self.app = QApplication(sys.argv)
        self.app.setStyleSheet(qdarktheme.load_stylesheet())
        super(RootWindow, self).__init__()
        self.conf = conf
        self.rc = None
        self.kb = None
        logging.basicConfig(
            level=getattr(logging, self.conf.c["canvas"]["logging"].upper(), 2)
        )
        self.logger = logging.getLogger(__name__)

        self.setWindowFlags(QtCore.Qt.WindowType.FramelessWindowHint)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_Hover, True)
        # self.mainWindow = QMainWindow(self)
        # self.mainWindow.setWindowFlags(QtCore.Qt.FramelessWindowHint)

        self.setAutoFillBackground(True)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_TranslucentBackground, True)
        self.setWindowTitle("SmartWHEEL")
        self.setWindowIcon(QIcon("logo.png"))
        self.setGeometry(*self.conf.c["window"]["geometry"])
        self.loadClasses()
        self.loadSerial()
        self.initUI()

        self.settings = SettingsWindow(
            os.path.join(self.conf["basedir"], "settings_registry", "config.json"),
            weakref.ref(self),
            weakref.ref(self.conf),
            self.conf["basedir"],
        )
        # self.settings.show()

        self.show()
        # self.qp = QPainter(self)

    def initUI(self):
        self.dock = QDockWidget()
        self.dock.setMaximumWidth(300)
        self.dock.setMaximumHeight(300)
        self.settingsButton = QPushButton("Settings", default=False, autoDefault=False)
        self.settingsButton.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.dock.setWidget(self.settingsButton)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.BottomDockWidgetArea, self.dock)
        self.dock.hide()

        self.settingsButton.clicked.connect(self.openSettings)

        self.installEventFilter(self)

        # self.drawBlur()

    @pyqtSlot()
    def openSettings(self):
        self.settings.show()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.Type.HoverEnter:
            self.dock.show()
        elif event.type() == QEvent.Type.HoverLeave:
            self.dock.hide()
        return super(RootWindow, self).eventFilter(obj, event)

    def drawBlur(self):
        self.blur_widget = QWidget(self)
        self.blur_widget.setGeometry(0, 0, *self.conf.c["window"]["geometry"][2:])
        self.blur_widget.setStyleSheet(
            "border-radius: {}px;\
                                       border: 0px solid white;\
                                       background-color: rgba(0, 0, 0, 0.5);".format(
                self.conf.c["window"]["geometry"][2] / 6
            )
        )
        # self.blur_widget.
        self.blur_filter = QGraphicsBlurEffect()
        self.blur_filter.setBlurRadius(15)
        os.system(
            "xprop -f _KDE_NET_WM_BLUR_BEHIND_REGION 32c -set _KDE_NET_WM_BLUR_BEHIND_REGION 0 -id "
            + str(int(self.winId()))
        )
        # self.blur_widget.setGraphicsEffect(self.blur_filter)
        self.blur_widget.show()

    def paintEvent(self, event):
        self.qp = QPainter(self)
        self.qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        # self.qp.begin(self)
        self.draw()
        self.qp.end()

    def keyPressEvent(self, event):
        """
        Read keypresses from Qt and pass them to serialpipe
        """
        if self.serialModules.get("serialpipe.keyboard") is None:  # not initialized yet
            return

        self.serialModules["serialpipe.keyboard"].handleKeypress(event)

    def keyPressEventLegacy(self, event):
        """
        Old keypress event, dead code
        """
        if event.key() == QtCore.Qt.Key_W:
            self.conf.c["canvas"]["modules"][0]["class"].processKey(True)
        elif event.key() == QtCore.Qt.Key_A:
            self.conf.c["canvas"]["modules"][0]["class"].processKey(False)
        elif event.key() == QtCore.Qt.Key_Up:
            self.rc.ae.action("scrollUp")
        elif event.key() == QtCore.Qt.Key_Down:
            self.rc.ae.action("scrollDown")
        elif event.key() == QtCore.Qt.Key_Return:
            self.rc.ae.action("keyAction1")
        elif event.key() == QtCore.Qt.Key_Q:
            self.hide()
            QtCore.QTimer.singleShot(1000, self.show)  # Hide example
        elif event.key() == QtCore.Qt.Key_Escape:
            sys.exit(0)
        self.update()

    def loadSerial(self):
        """
        Init main serial connections
        """
        # TODO add serial.tools.miniterm in settings

        self.serialModules = {}
        self.serialModulesNames = []

        for i in self.conf.c["canvas"]["serialModulesLoad"]:
            mod_name = self.conf.c["canvas"]["serialModules"][i]["name"]
            mod = importlib.import_module("smartwheel." + mod_name)

            try:
                cls = mod.SConn(
                    os.path.join(
                        self.conf["config_dir"],
                        self.conf.c["canvas"]["serialModules"][i]["config"],
                    ),
                    self.rc.ae.callAction,
                )
                cls.start()
                self.serialModules[mod_name] = cls
                self.serialModulesNames.append(mod_name)
            except BaseException as e:
                self.logger.error("Failed to load " + mod_name + ": ", e)

    def loadClasses(self):
        self.rc = RootCanvas(
            self.conf.c_canvas, self.conf.launch_config["config_dir"], self.update
        )

        QGuiApplication.instance().aboutToQuit.connect(self.rc.killThreads)

    def draw(self):
        self.rc.draw(self.qp)


def main():
    dirpath = os.path.dirname(os.path.realpath(__file__))

    if not os.path.exists(dirpath):
        print(
            "Please check that configuration files are present in the installation directory"
        )

    with open(os.path.join(dirpath, "launch.json"), "r") as f:
        launch_config = json.load(f)

    launch_config["config_dir"] = os.path.join(dirpath, launch_config["config_dir"])

    main_conf_path = os.path.join(launch_config["config_dir"], "config.json")
    conf = WConfig(main_conf_path, launch_config)
    conf["config_dir"] = launch_config["config_dir"]
    conf["basedir"] = dirpath
    conf.c_canvas["basedir"] = conf["basedir"]
    root = RootWindow(conf)
    sys.exit(root.app.exec())


if __name__ == "__main__":
    main()



import logging

from PyQt6.QtWidgets import QListView

from smartwheel.settings_handlers.base import BaseHandler


class SerialHandler(BaseHandler):
    def __init__(self, value_getter, value_setter, parent_obj=None):
        super(SerialHandler, self).__init__(value_getter, value_setter, parent_obj)
        self.logger = logging.getLogger(__name__)

    def initElem(self, elem):
        l_view = QListView()



# from PyQt6.QtGui import *
import importlib
import json
import logging
import os
import weakref
from queue import LifoQueue

from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtWidgets import (
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from smartwheel import common


class SettingsWindow(QWidget):
    def __init__(self, config_file, main_class, conf_class, basedir, parent=None):
        """
        Init SettingsWindow

        Parameters
        ==========
        config_file
            Path to settings registry
        main_class
            Weakref to RootWindow object
        conf_class
            Weakref to WConfig object
        """
        super(SettingsWindow, self).__init__(parent)
        self.setWindowTitle("Settings")
        self.hook = None
        self.conf = None
        self.handlers_conf = None
        self.handlers = None
        self.basedir = basedir
        self.settings = {}
        self.external_reg = {}
        self.logger = logging.getLogger(__name__)
        self.loadConfig(config_file)
        self.setConfigHook(main_class, conf_class)
        self.loadSettingsHandlers(
            os.path.join(self.basedir, self.conf["settings_handlers_dir"])
        )
        self.preset_tabs = {}
        self.conf["presets"] = {}
        self.linked_widgets = {}
        self.presets_queue = LifoQueue()
        self.presets_update_queue = []
        self.isLoaded = False

        self.initLayout()

    def loadConfig(self, config_file):
        """
        Import settings registry

        Parameters
        ==========
        config_file
            Settings registry config.json file
        """
        with open(config_file, "r") as f:
            self.conf = json.load(f)

        if self.conf.get("tabs") is not None:
            for i in range(len(self.conf["tabs"])):
                with open(
                    os.path.join(
                        self.basedir,
                        "settings_registry",
                        self.conf["tabs"][i]["config"],
                    ),
                    "r",
                ) as f:
                    self.conf["tabs"][i]["conf"] = json.load(f)

            for key, value in self.conf["external"].items():
                with open(
                    os.path.join(
                        self.basedir, "settings_registry", "external", key + ".json"
                    ),
                    "r",
                ) as f:
                    self.external_reg[key] = json.load(f)
                self.external_reg[key]["extra"] = self.conf["external"][key]

    def setConfigHook(self, main_class, conf):
        """
        Set hooks for the settings class in order to load/save global settings

        Parameters
        ==========
        main_class
            Weakref to RootWindow object
        conf
            Weakref to WConfig object
        """
        self.main_class = main_class  # weakref to the main class
        self.confClass = conf  # weakref to wconfig
        self.settings["main"] = conf().c
        self.settings["canvas"] = main_class().rc.conf
        self.settings["common"] = main_class().rc.common_config
        self.settings["settings"] = self.conf
        self.settings["actionengine"] = main_class().rc.ae.conf

        # Parsing serial modules
        self.settings["serial"] = {}
        serial = main_class().serialModules

        for name in main_class().serialModulesNames:
            if hasattr(serial[name], "conf"):
                self.settings["serial"][name] = serial[name].conf
            else:
                self.logger.error("serialpipe." + name + " has no conf attribute")

        # Parsing canvas section modules
        self.settings["modules"] = {}
        main_modules = main_class().rc.conf["modules"]
        wheel_modules = main_class().rc.wheel_modules

        for i, modules in enumerate([main_modules, wheel_modules]):
            for mod in modules:
                if mod.get("class") is not None:
                    if hasattr(mod["class"], "conf"):
                        self.settings[mod["name"]] = mod["class"].conf
                    else:
                        self.logger.error(mod["name"] + " has no conf attribute")
                else:
                    if i == 1:
                        self.logger.error(mod["name"] + " has no class attribute")

    def loadSettingsHandlers(self, handlers_dir):
        """
        Init settings handlers from the directory

        Parameters
        ==========
        handlers_dir
            Directory containing settings handlers, must contain config.json file
        """
        s_config = os.path.join(handlers_dir, "config.json")
        if not os.path.exists(s_config):
            self.logger.error("Missing " + s_config + " file")
            os.exit(1)

        with open(s_config, "r") as f:
            self.handlers_conf = json.load(f)

        self.handlers = {}
        for mod_name in self.handlers_conf["handlers_modules"]:
            handler = importlib.import_module(
                "smartwheel." + self.conf["settings_handlers_dir"] + "." + mod_name
            )
            h_dict = handler.handlers
            for k in h_dict:
                if self.handlers.get(k) is not None:
                    self.logger.warning(
                        "Setting handler " + k + " is defined twice. Overriding"
                    )
                self.handlers[k] = h_dict[k](
                    self.getValue, self.setValue, weakref.ref(self)
                )

    def getValue(self, module, prop, index=None):
        """
        Get property from the application
        Returns (True, value) if found, (False, None) otherwise

        Parameters
        ==========
        module
            Module name (in self.settings)
        prop
            Property keys, separated by `.`
        index
            If not None, the index in the property array. If an array, then it's duplicated at specified indices
        """
        if self.settings.get(module) is None:
            self.logger.error("Could not get value: no module " + module)
            return False, None

        props = prop.split(".")
        cur_prop = self.settings[module]
        for p in props:
            if cur_prop.get(p) is None:
                self.logger.error(
                    "Could not get value: no property " + module + "." + prop
                )
                return False, None
            else:
                cur_prop = cur_prop[p]

        if index is None:
            return True, cur_prop
        elif type(index) == list:
            return True, cur_prop[index[0]]
        else:
            return True, cur_prop[index]

    def dictWalk(self, d, props, value, index=None, _i=0):
        """
        Recursively walk in nested dicts and apply value

        Returns True if new value is different

        Parameters
        ==========
        d
            Target dict
        props
            List of keys
        value
            Value to apply
        index
            If not None, the index in the property array. If an array, then it's duplicated at specified indices
        _i
            (Private) Recursion depth
        """

        key = "_" + str(_i)

        if _i != 0:
            key = "_" + str(_i - 1)

        # We create the wrapper if needed
        if d.get(key) is None:
            w = {key: d}
        else:
            w = d

        if len(props) == 1:
            if index is None:
                new = not w[key].get(props[0]) == value
                w[key][props[0]] = value
            elif type(index) == list:
                new = False
                for i in index:
                    new = new or (not w[key][props[0]][i] == value)
                    w[key][props[0]][i] = value
            else:
                new = not w[key][props[0]][index] == value
                w[key][props[0]][index] = value
            return new

        # Python reference hack, we use a mutable type as a wrapper
        wrapper = {"_" + str(_i): w[key][props[0]]}

        return self.dictWalk(wrapper, props[1:], value, index, _i + 1)

    def setValue(self, obj=None, value=None, module=None, prop=None, index=None):
        """
        Set property from the application.
        Module, prop and index arguments may be fetched from the object properties (passing only obj and value) or passed directly (module, prop, index (optional) and value)

        Parameters
        ==========
        obj
            (Optional) QObject (widget) with `module`, `prop` (and `index`) properties
        value
            (Optional) Value to set
        module
            (Optional) Module name (in self.settings)
        prop
            (Optional) Property keys, separated by `.`
        index
            (Optional) If not None, the index in the property array. If an array, then it's duplicated at specified indices
        """

        if obj is not None:
            module = obj.property("widmodule")
            prop = obj.property("prop")
            index = obj.property("index")

        if module is None:
            self.logger.error("Could not obtain module value")

        if self.settings.get(module) is None:
            self.logger.error("Could not get value: no module " + str(module))
            return

        props = prop.split(".")

        if self.dictWalk(self.settings[module], props, value, index):
            if self.isLoaded:
                common.config_manager.updated.emit(props[-1:][0])
            else:
                self.presets_update_queue.append(props[-1:][0])

            self.main_class().update()

    def savePreset(self, index, name, title, filepath):
        """
        Save the specified preset

        Parameters
        ==========
        index
            Settings tab index
        name
            Preset internal name
        title
            Preset title
        filepath
            Path to the preset json file (may not exist)
        """
        preset = {"name": name, "title": title, "props": {}}

        for key, _ in self.preset_tabs[index].items():
            elem, handler = self.preset_tabs[index][key]
            prop = elem().property("preset_property")
            value = handler.fetchValue(elem())

            if value is None:
                continue

            preset["props"][prop] = value

        self.conf["presets"][str(index)][name] = preset

        with open(filepath, "w") as f:
            json.dump(preset, f, indent=4)

        return True, preset

    def loadPreset(self, index, name):
        """
        Load the preset from the handler

        Parameters
        ==========
        index
            Settings tab index
        name
            Preset internal name
        """
        preset = self.conf["presets"][str(index)][name]

        for key, value in preset["props"].items():
            p_elem = self.preset_tabs[index].get(key)
            if p_elem is None:
                if not self.isLoaded:
                    self.presets_queue.put((index, name))
                else:
                    self.logger.warning(
                        "Could not find "
                        + key
                        + " widget from preset "
                        + preset["title"]
                    )
                continue

            elem, handler = p_elem

            ok = handler.updateValue(elem(), value)
            if not ok:
                self.logger.warning(
                    "Could not set "
                    + key
                    + " widget value from preset "
                    + preset["title"]
                )

    @pyqtSlot(str)
    def showLinkedWidgets(self, text):
        """
        Show/hide linked widgets of the external registry selector. Intended to work with QComboBox

        Parameters
        ==========
        index
            Index of selected element
        """
        caller = self.sender()

        linked_to = caller.property("registriesName")

        if linked_to is None:
            self.logger.warning(
                "External registries widget does not have registries name"
            )

        self._showWidgets(linked_to, text)

    def _showWidgets(self, linked_to, text):
        for key, pair in self.linked_widgets[linked_to].items():
            for p in pair:
                if key == text:
                    p[0]().setRowVisible(p[1], True)
                else:
                    p[0]().setRowVisible(p[1], False)

    def processItem(self, elem, index, form, tab, registriesName=None):
        if elem.get("name") is not None:
            label = QLabel(elem["name"])
        else:
            label = None

        wid = None

        elem["tab_index"] = index  # storing the index of the current tab

        if self.handlers.get(elem["type"]) is None:
            self.logger.error(
                "Could not find the handler for type "
                + elem["type"]
                + " ("
                + elem["module"]
                + "."
                + elem["prop"]
                + ")"
            )
        else:
            wid = self.handlers[elem["type"]].initElem(elem)

        if wid is not None:
            prop_preset = ""

            if elem.get("module") is not None:
                wid.setProperty("widmodule", elem["module"])
                prop_preset += elem["module"]

            if elem.get("prop") is not None:
                wid.setProperty("prop", elem["prop"])
                prop_preset += "." + elem["prop"]

            if elem.get("index") is not None:
                wid.setProperty("index", elem["index"])
                prop_preset += "." + str(elem["index"])

            if (
                tab["conf"].get("enable_presets", False)
                and elem.get("preset", False)
                and not prop_preset == ""
            ):
                wid.setProperty("preset_property", prop_preset)

                self.preset_tabs[index][prop_preset] = (
                    weakref.ref(wid),
                    self.handlers[elem["type"]],
                )

            if registriesName is not None:
                ok = self.handlers[elem["type"]].linkElem(wid, registriesName)
                if not ok:
                    self.logger.warning(
                        "Element type "
                        + elem["type"]
                        + " does not support linking extra registries"
                        + " ("
                        + elem["module"]
                        + "."
                        + elem["prop"]
                        + ")"
                    )

            widWrapper = QHBoxLayout()

            if label is not None:
                wid.setMinimumWidth(self.conf["fieldWidth"])
                wid.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum)
                spacer = QSpacerItem(
                    40,
                    20,
                    QSizePolicy.Policy.Expanding,
                    QSizePolicy.Policy.Minimum,
                )
                widWrapper.addSpacerItem(spacer)

            widWrapper.addWidget(wid)

            if label is not None:
                form.addRow(label, widWrapper)
            else:
                form.addRow(widWrapper)

        return form.rowCount() - 1, wid

    def initTab(self, index):
        """
        Parse registry and generate elements

        Parameters
        ==========
        index
            Tab index
        """
        tab = self.conf["tabs"][index]

        if tab["conf"].get("enable_presets", False):
            self.preset_tabs[index] = {}

        scroll = QScrollArea()
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        scroll.setWidgetResizable(True)
        layout = QVBoxLayout()
        wrapper = QWidget()
        wrapper.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        for elem_group in tab["conf"]["items"]:
            form = QFormLayout()
            group = QGroupBox(elem_group["name"])

            items = elem_group["options"]

            for i, elem in enumerate(items):
                if elem.get("external") is not None:
                    if type(elem["external"]) == dict:
                        ok, ex_registries = self.getValue(
                            module=elem["external"]["module"],
                            prop=elem["external"]["prop"],
                            index=elem["external"].get("index"),
                        )

                        if not ok:
                            self.logger.error(
                                "Could not find external registries in " + elem["name"]
                            )
                            continue

                        if not type(ex_registries) == list:
                            ex_registries = [ex_registries]
                    else:
                        ex_registries = elem["external"]

                    if elem["external"].get("pickerName") is None:
                        self.logger.error(
                            "No registry specified for element " + elem["name"]
                        )
                        continue

                    registries_name = elem["external"]["pickerName"]

                    _, controller = self.processItem(
                        elem, index, form, tab, registriesName=registries_name
                    )

                    self.linked_widgets[registries_name] = {}

                    for reg_name in ex_registries:
                        if self.external_reg.get(reg_name) is None:
                            self.logger.warning(
                                "No such external registry: " + reg_name
                            )
                            continue

                        exr = self.external_reg[reg_name]

                        if exr.get("items") is None:
                            self.logger.error(
                                "External registry " + reg_name + " does not have items"
                            )
                            continue

                        if not type(exr["items"]) == list:
                            self.logger.warning(
                                "External registry " + reg_name + " is not a list"
                            )
                            continue

                        self.linked_widgets[registries_name][
                            exr["extra"]["linkedCombo"]
                        ] = []

                        for reg in exr["items"]:
                            form_row, _ = self.processItem(reg, index, form, tab)
                            self.linked_widgets[registries_name][
                                exr["extra"]["linkedCombo"]
                            ].append((weakref.ref(form), form_row))

                        self._showWidgets(registries_name, controller.currentText())

                else:
                    self.processItem(elem, index, form, tab)

            group.setLayout(form)
            layout.addWidget(group)

        wrapper.setLayout(layout)
        scroll.setWidget(wrapper)

        return scroll

    def initLayout(self):
        """
        Generate layout
        """
        baseLayout = QVBoxLayout(self)

        tabWidget = QTabWidget()
        for i in range(len(self.conf["tabs"])):
            scroll = self.initTab(i)
            tabWidget.addTab(scroll, self.conf["tabs"][i]["name"])

        baseLayout.addWidget(tabWidget)

        bottomPanel = QHBoxLayout()

        okButton = QPushButton("OK")
        okButton.clicked.connect(common.config_manager.saveConfig)
        okButton.clicked.connect(self.close)

        applyButton = QPushButton("Apply")
        applyButton.clicked.connect(common.config_manager.saveConfig)

        cancelButton = QPushButton("Cancel")
        cancelButton.clicked.connect(self.close)
        spacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        bottomPanel.addSpacerItem(spacer)
        bottomPanel.addWidget(cancelButton)
        bottomPanel.addWidget(applyButton)
        bottomPanel.addWidget(okButton)

        baseLayout.addLayout(bottomPanel)

        self.setLayout(baseLayout)

        while not self.presets_queue.empty():
            self.loadPreset(*self.presets_queue.get())

        self.isLoaded = True

        # updating properties
        if self.presets_update_queue:
            common.config_manager.batchUpdate.emit(self.presets_update_queue)



import logging

import serial
from PyQt6.QtCore import *

from smartwheel import config
from smartwheel.serialpipe.base import ConnPipe


class SConn(ConnPipe):
    """Default serial connection interface (compatible with pico)"""

    def __init__(self, config_file, call_signal):
        super().__init__()
        self.conf = None
        self.config_file = config_file
        self.call = call_signal
        self.logger = logging.getLogger(__name__)
        self.loadConfig()

    def loadConfig(self):
        self.conf = config.Config(self.config_file, self.logger)
        self.conf.loadConfig()

    def serialCall(self, string):
        """
        Parse serial call from string

        Parameters
        ----------
        string
            Command from serial
        """
        # l = string.split()
        string = string.decode("utf-8").strip()
        print(string)
        if string == "":
            return
        cmd = []
        for b in self.conf["binds"]:
            for c in b["commands"]:
                if c["string"] == string:
                    cmd.append((b, c))
        if cmd != []:
            for c in cmd:
                if c != []:
                    self.call.emit(c)

    def run(self):
        """
        Reads for serial data from device and executes actions
        """
        # TODO move to another module
        tm = None
        if self.conf["useTimeout"]:
            tm = self.conf["timeout"]
        try:
            with serial.Serial(
                self.conf["device"], self.conf["baudRate"], timeout=tm
            ) as s:
                while self.isRunning() and s.is_open:
                    self.serialCall(s.readline())
        except BaseException as e:
            self.logger.error(e)
            return



from smartwheel import config
from smartwheel.ui.base import BaseUIElem


class UIElem(BaseUIElem):
    def __init__(self, config_file, WConfig):
        super().__init__()
        self.is_folder = True
        self.config_file = config_file
        self.loadConfig()
        self.conf.c = {**self.conf, **WConfig}
        self.icon_path = self.conf.get("iconPath", None)
        self.modules = self.conf["modules"]
        self.wrapper_pointer = None

    def loadConfig(self):
        self.conf = config.Config(self.config_file)
        self.conf.loadConfig()

    def processKey(self, event):
        event["canvas"]().reloadWheelModules(
            False, self.wrapper_pointer()
        )  # call canvas module

    def draw(self, qp, offset=None):
        pass



from smartwheel.actions.wheel.base import BaseWheelAction


class WheelAction(BaseWheelAction):
    def __init__(self):
        self.type = "wheel"

    def run(self, action, canvas, ae):
        wheel = canvas().conf["modules"][0]["class"]
        if action == "wheelUp":
            wheel.processKey(True)
        elif action == "wheelDown":
            wheel.processKey(False)
        elif action == "wheelSelect":
            wheel.selectModule()
        elif action == "wheelOpen":
            wheel.openWheel()
        elif action == "wheelQuickUp":
            wheel.quickSwitch(True)
        elif action == "wheelQuickDown":
            wheel.quickSwitch(False)
